
# coding: UTF-8

##
## @file PmxLib.py
## @brief PMXで使用する関数や定義をひとまとまりにしたものです

## @version V1.0.2.0
## @author Kondo Kagaku Co.,Ltd.
##          - kedtn(T,Nobuhara)
##          - hn.kondo
##          - gotanda
## @date 2024/07
## @copyright &copy; Kondo Kagaku Co.,Ltd. 2024
## @copyright [MIT License](http://opensource.org/licenses/mit-license.php)

## @mainpage PmxPythonLibの概要
## @section overview 概要
## このライブラリは近藤科学社製PMXサーボモータをpythonで制御させるためのライブラリです<br>
## <A HREF= "https://kondo-robot.com/product-category/servomotor/pmx">PMXサーボモータ</A> <br>
## 使い方および詳細は、下記弊社HPをご覧ください。<br>
## <A HREF="http://kondo-robot.com/">http://kondo-robot.com/</A><br>
## PMXの通信方法等は<A HREF= https://kondo-robot.com/faq/pmx-servo-series-online-manual>PMXのオンラインマニュアル</A>をご覧ください
## 不具合等ありましたら、弊社HPを参照にご連絡ください。<br>
## @section changelog 更新履歴
## - 2023/10 : 1.0.0.0 
##     - First Release
## - 2023/11 : V1.0.0.1
##     - 軽微なバグを修正
## - 2024/02 : V1.0.1.0
##      - PMXファームウェアバージョンV.1.0.1.0に対応する下記の機能を追加しました。
##      - RamAddrListにショートブレーキ設定(530~)および目標指令値(700番台)のアドレスを追加しました
##      - MotorWriteコマンドを使用せずMotorのON/OFFが設定する「setTorqueSwitch」関数を追加しました
##      - setControlModeのwriteOptが反映されていなかったので修正しました
## - 2024/07 : V1.0.2.0
##      - PMXファームウェアバージョンV.1.1.0.0に対応する下記の機能を追加しました。
##      - 制御モードの組み合わせが追加されました
##      - RamAddrListに制御ゲイン、プリセットゲイン、LED点灯モードのアドレスを追加しました
##      - getPresetAllPidGainsほか制御ゲイン、プリセットゲイン関係の関数を追加しました
##      - LED点灯モードに対応するgetLedMode、setLedMode関数を追加しました
##      - エラーリストに実行エラーを追加しました
##      



import copy

from dataclasses import dataclass

class PmxLib():
    """
    @brief PMXで使用する関数や定義をひとまとまりにしたものです
    @details
    * PMXの固有値の定義
    * PMXのコマンドの生成、チェックおよび送受信(送信部分は外部依存)
    * PMXで必要だと思われる機能の関数化
    @details をしています
    
    @warning 通信部分は外部に定義しますのでこのclassを派生して送信に必要な関数を追加してください。
    """

    @dataclass(frozen=True)
    class RamAddrList:
        '''
        PMXサーボモータに格納されるメモリ内の先頭アドレスを定義します
        '''
        ## 位置制御のPゲイン
        PositionKp = 0
        ## 位置制御のIゲイン
        PositionKi = 4
        ## 位置制御のDゲイン
        PositionKd = 8
        ## 位置制御のストレッチ
        PositionStrc = 12
        
        ## 速度制御のPゲイン
        SpeedKp = 16
        ## 速度制御のIゲイン
        SpeedKi = 20
        ## 速度制御のDゲイン
        SpeedKd = 24
        
        ## 電流制御のPゲイン
        CurrentKp = 32
        ## 電流制御のIゲイン
        CurrentKi = 36
        ## 電流制御のDゲイン
        CurrentKd = 40
        
        ## トルク制御のPゲイン
        TorqueKp = 48
        ## トルク制御のIゲイン
        TorqueKi = 52
        ## トルク制御のDゲイン
        TorqueKd = 56
        
        ## 位置制御　不感帯
        PositionDeadBand = 64
        ## 速度制御　不感帯
        SpeedDeadBand = 66
        ## 電流制御　不感帯
        CurrentDeadBand = 68
        ## トルク制御　不感帯
        TorqueDeadBand = 70
        
        ## 中央値オフセット
        CenterOffset = 72
        ## クローン/リバース
        CloneReverse = 74

        ## 入力電圧最小値
        MinVoltageLimit = 76
        ## 入力電圧最小時の出力％値
        MinVoltageLimitPower = 78
        ## 入力電圧最大値
        MaxVoltageLimit = 80
        ## 入力電圧最大時の出力％値
        MaxVoltageLimitPower = 82
        ## モータ消費電流最大値
        CurrentLimit = 84
        ## モータ消費電流最大時の出力％値
        CurrentLimitPower = 86
        ## モータ温度最大値
        MotorTempLimit = 88
        ## モータ温度最大時の出力％値
        MotorTempLimitPower = 90
        ## CPU温度最大値
        CpuTempLimit = 92
        ## CPU温度最大時の出力％値
        CpuTempLimitPower: int = 94

        ## CW方向最大角値
        CwPositionLimit = 96
        ## CW方向最大角閾値外時の出力％値
        CwPositionLimitPower = 98
        ## CCW方向最大角値
        CcwPositionLimit = 100
        ## CCW方向最大角閾値外時の出力％値
        CcwPositionLimitPower = 102
        ## 最大速度指令値
        MaxGoalSpeed = 104
        ## 最大電流指令値
        MaxGoalCurrent = 106
        ## 最大動推定トルク指令値
        MaxGoalTorque = 108
        ## モータ出力制限％値
        TotalPowerRate = 110
        ## ロック時間
        LockDetectTime = 112
        ## ロックと認識される出力割合
        LockThresholdPower = 114
        ## ロック時間の出力％値
        LockDetectOutputPower = 116

        ## 位置制御ゲインプリセット(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        presetPosAddr = 118
        ## 速度制御ゲインプリセット(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        presetSpdAddr = 119
        ## 電流制御ゲインプリセット(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        presetCurAddr = 120
        ## トルク制御ゲインプリセット(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        presetTrqAddr = 121

        ## 位置制御ゲインKp2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        PositionKp2 = 124
        ## 位置制御ゲインKi2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        PositionKi2 = 128
        ## 位置制御ゲインKd2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        PositionKd2 = 132
        ## 位置制御ストレッチ2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        PositionStrc2 = 136
        ## 速度制御ゲインKp2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        SpeedKp2 = 140
        ## 速度制御ゲインKi2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        SpeedKi2 = 144
        ## 速度制御ゲインKd2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        SpeedKd2 = 148
        ## 電流制御ゲインKp2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        CurrentKp2 = 156
        ## 電流制御ゲインKi2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        CurrentKi2 = 160
        ## 電流制御ゲインKd2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        CurrentKd2 = 164
        ## トルク制御ゲインKp2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        TorqueKp2 = 172
        ## トルク制御ゲインKi2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        TorqueKi2 = 176
        ## トルク制御ゲインKd2(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        TorqueKd2 = 180
        ## 位置制御ゲインKp3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        PositionKp3 = 188
        ## 位置制御ゲインKi3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        PositionKi3 = 192
        ## 位置制御ゲインKd3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        PositionKd3 = 196
        ## 位置制御ストレッチ3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        PositionStrc3 = 200
        ## 速度制御ゲインKp3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        SpeedKp3 = 204
        ## 速度制御ゲインKi3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        SpeedKi3 = 208
        ## 速度制御ゲインKd3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        SpeedKd3 = 212
        ## 電流制御ゲインKp3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        CurrentKp3 = 220
        ## 電流制御ゲインKi3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        CurrentKi3 = 224
        ## 電流制御ゲインKd3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        CurrentKd3 = 228
        ## トルク制御ゲインKp3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        TorqueKp3 = 236
        ## トルク制御ゲインKi3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        TorqueKi3 = 240
        ## トルク制御ゲインKd3(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        TorqueKd3 = 244

        ## 現在位置
        NowPosition = 300
        ## 現在速度
        NowSpeed = 302
        ## 現在電流
        NowCurrent = 304
        ## 出力推定トルク
        NowTorque = 306
        ## 現在のPWM割合
        NowPwm = 308
        ## モータ温度
        MotorTemp = 310
        ## CPU温度
        CPUTemp = 312
        ## 入力電圧
        InputVoltage = 314
        ## 現在の補間時間
        TrajectoryTime = 316
        ## エンコーダの現在値
        EncoderValue = 318
                
        ## エラーステータス
        ErrorStatus = 400
        ## システムエラー
        ErrorSystem = 401
        ## モータエラー
        ErrorMotor = 402
        ## RAMアクセスエラー時のアドレス
        ErrorRamAccess = 404
        
        ## トルクスイッチ
        TorqueSwitch = 500
        ## 制御モード
        ControlMode = 501
        ## 応答データフラグ
        MotorReceiveData = 502
        ## 補間制御起動生成タイプ指定
        Trajectory = 503
        
        ## 電流制御時のショートブレーキ指定(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        ShortBrakeCurrent = 530
        ## トルク制御時のショートブレーキ指定(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        ShortBrakeTorque = 531
        ## PWM制御時のショートブレーキ指定(PMXのV1.0.1.x～)
        ## @remark PMXのV1.0.1.xから有効
        ShortBrakePWM = 532

        ## LED点灯モード(PMXのV1.1.0.x～)
        ## @remark PMXのV1.1.0.xから有効
        LedMode = 533

        ## 中央値オフセット　最小値
        CenterOffsetMinRange = 600
        ## 中央値オフセット　最大値
        CenterOffsetMaxRange = 602
        ## 入力電圧最小制限値　最小値
        MinVoltageMinRange = 604
        ## 入力電圧最小制限値　最大値
        MinVoltageMaxRange = 606
        ## 入力電圧最大制限値　最小値
        MaxVoltageMinRange = 608
        ## 入力電圧最大制限値　最大値
        MaxVoltageMaxRange = 610
        ## フェールセーフ電圧　最小値
        FailSafeVoltageMinRange = 612
        ## フェールセーフ電圧　最大値
        FailSafeVoltageMaxRange = 614
        ## モータ消費電流設定値　最小値
        CurrentMinRange = 616
        ## モータ消費電流設定値　最大値
        CurrentMaxRange = 618
        ## モータ温度設定値　最小値
        MotorTempMinRange = 620
        ## モータ温度設定値　最大値
        MotorTempMaxRange = 622
        ## CPU温度設定値　最小値
        CpuTempMinRange = 624
        ## CPU温度設定値　最大値
        CpuTempMaxRange = 626
        ## CW方向最大値　最小値
        CwPositionMinRange = 628
        ## CW方向最大値　最大値
        CwPositionMaxRange = 630
        ## CCW方向最大値　最小値
        CcwPositionMinRange = 632
        ## CCW方向最大値　最大値
        CcwPositionMaxRange = 634
        ## 最大速度設定値　最小値
        MaxGoalSpeedMinRange = 636
        ## 最大速度設定値　最大値
        MaxGoalSpeedMaxRange = 638
        ## 最大電流設定値　最小値
        MaxGoalCurrentMinRange = 640
        ## 最大電流設定値　最大値
        MaxGoalCurrentMaxRange = 642
        ## 最大トルク設定値　最小値
        MaxGoalTorqueMinRange = 644
        ## 最大トルク設定値　最大値
        MaxGoalTorqueMaxRange = 646
        
        ## 目標指令値1
        ## @remark PMXのV1.0.1.xから有効
        GoalCommandValue1 = 700
        ## 目標指令値2
        ## @remark PMXのV1.0.1.xから有効
        GoalCommandValue2 = 702
        ## 目標指令値3
        ## @remark PMXのV1.0.1.xから有効
        GoalCommandValue3 = 704


    @dataclass(frozen=True)
    class ControlMode:
        """
        位置や速度制御等の切り替えの値を定義します
        """
        ## 位置制御
        Position = 0x01
        ## 速度制御
        Speed = 0x02
        ## 位置/速度制御
        ## @remark PMXのV1.1.0.xから有効
        PositionSpeed = 0x03
        ## 電流制御
        Current = 0x04
        ## 位置/電流制御
        PositionCurrent = 0x05
        ## 速度/電流制御
        ## @remark PMXのV1.1.0.xから有効
        SpeedCurrent = 0x06
        ## 位置/速度/電流制御
        PositionSpeedCurrent = 0x07
        ## トルク制御
        Torque = 0x08
        ## 位置/トルク制御
        PositionTorque = 0x09
        ## 速度/トルク制御
        ## @remark PMXのV1.1.0.xから有効
        SpeedTorque = 0x0A
        ## 位置/速度/トルク制御
        ## @remark PMXのV1.1.0.xから有効
        PositionSpeedTorque = 0x0B
        ## PWM制御
        PWM = 0x10
        ## 時間(単体制御はないので定義のみ)
        Time = 0x20
        ## 位置/時間制御
        PositionTime = 0x21
        ## 位置/電流/時間制御
        PositionCurrentTime = 0x25
        ## 位置/トルク/時間制御
        PositionTorqueTime = 0x29
        
    
    @dataclass(frozen=True)
    class TrajectoryType:
        """
        PMXの時間制御で使用する加減速具合を設定します
        """
        ## 均等補間
        Even = 0x01
        ## 5次多項式補間
        FifthPoly = 0x05
        ## エラー
        Error = 0xFF

        
    @dataclass(frozen=True)
    class EditBaudrate:
        """
        PMXで設定する通信速度の値を定義します
        """
        ## 57600bps
        _57600 = 0x00
        ## 115200bps
        _115200 = 0x01
        ## 625000bps
        _625000 = 0x02
        ## 1Mbps
        _1000000 = 0x03
        ## 1.25Mbps
        _1250000 = 0x04
        ## 1.5Mbps
        _1500000 = 0x05
        ## 2Mbps
        _2000000 = 0x06
        ## 3Mbps
        _3000000 = 0x07


    @dataclass(frozen=True)
    class EditParity:
        """
        PMXで設定するパリティの値を定義します
        """
        ParityNone = 0x00
        Odd = 0x01
        Even = 0x02


    @dataclass(frozen=True)
    class SendCmd:
        ''' 
        送信コマンド一覧
        '''
        ## RAM上のデータを読み取る
        MemREAD = 0xa0
        ## RAM上へデータを書き込む
        MemWRITE = 0xa1
        ## フラッシュメモリーに保存されているデータをRAMへ展開する
        LOAD = 0xa2
        ## RAMのデータをフラッシュメモリーに保存する
        SAVE = 0xa3
        ## ポジション指定コマンド
        MotorREAD = 0xa4
        ## RAM上へデータを書き込む
        MotorWRITE = 0xa5
        ## aA9～BA空き
        ## IDシリアル番号等の読み込み
        SystemREAD = 0xbb
        ## ID・通信速度・パリティ書き込み
        SystemWRITE = 0xbc
        ## リセットコマンド
        ReBoot = 0xbd
        ## 工場出荷時設定へ戻す
        FactoryReset = 0xbe


    @dataclass(frozen=True)
    class BuffPter:
        '''
        送受信データ位置
        '''
        Header = 0
        Header1 = 1
        ID = 2
        Length = 3
        CMD = 4
        Option = 5
        Status = 5
        Data = 6


    @dataclass(frozen=True)
    class ReceiveDataOption:
        """
        MotorREAD等で使用する応答モードのリスト
        @details 複数のデータの返信をもらうには、全て足したものを使用します。
        @code
        
        # 位置と電流値を返す値
        returnData = PmxLib.ReceiveDataOption.Position + PmxLib.ReceiveDataOption.Current
        @endcode
        """
        
        NoReturn = 0x00
        Position = 0x01
        Speed = 0x02
        Current = 0x04
        Torque = 0x08
        Pwm = 0x10
        MotorTemp = 0x20
        CpuTemp = 0x40
        Voltage = 0x80
        Full = 0xFF


    @dataclass(frozen=True)
    class ComError:
        """
        データが返ってこないなどサーボモータ外に起因するエラー一覧の定義
        @details サーボモータ外に起因するエラー(通信タイムアウト等)のデータを定義しています
        @details PMXと通信を行い、結果を返すときに使用します。
        @note PMXと通信を行った際、PMXのstatusは1byteなので、頭にもう1byte付与しエラー値とします
        @note PMXの通信結果を返す場合関数は、このComErrorとモータの返信status値を付与(OR)した値を返します
        """
        
        ## 正常
        OK = 0
        ## 無応答/Timeoutエラー
        TimeOut = 0xFF00
        ## CRCエラー
        CrcError = 0xFE00
        ## 作られたデータのフォーマットに異常があった
        FormatError = 0xFD00
        ## 送信する際に異常があった
        SendError = 0xFC00
        ## 受信したデータに異常があった
        ReceiveError = 0xFB00
        ## MotorREADのデータ変換に異常があった
        MotorREADConvertError = 0xFA00
        
        ## 上記以外のエラー
        NG = 0x8000
        ## マスク用のデータ
        ErrorMask = 0xFF00


    @dataclass(frozen=True)
    class MinimumLength:
        ''' コマンドの最低サイズ値の定義 '''
        
        ## 送信コマンドの最低サイズ
        Send = 8  # Header(2),id(1),length(1),cmd(1),st(1),crc(2)
        
        ## 受信コマンドの最低サイズ
        Receive = 8  # Header(2),id(1),length(1),cmd(1),st(1),crc(2)


    @dataclass(frozen=True)
    class PmxStatusErrorList:
        """
        PMXのステータスの返信/エラー状態の値
        """
        ## System Error
        SystemError = 0x01
        ## Motor Error
        MotorError = 0x02
        ## Communication Error
        CommunicationError = 0x04
        ## Command Error
        CommandError = 0x08
        ## RAM Access Error
        RamAccessError = 0x10
        ## Mode Error
        ModeError = 0x20
        ## Data Error
        DataError = 0x40
        ## run Error
        RunError = 0x80
        

    @dataclass(frozen=True)
    class TorqueSwitchType:
        """ TorqueOnの値の定義 """

        ## MotorWriteのオプションで使用
        Control = 0x00
        
        ## トルクON 
        TorqueOn = 0x01
        ## フリーモード
        Free = 0x02 
        ## ショートブレーキモード
        Brake = 0x04
        ## Holdモード
        Hold = 0x08
        ## トルクON/OFFを抜き出すためのマスクする値
        Mask = 0x0F
        ## 返事なし
        # NoReply = 0x10
        ## エラー
        Error = 0xFF


    @dataclass(frozen=True)
    class CloneReverseType:
        """ CloneReverseの値の定義 """
        
        ## cloneOnの設定値
        Clone = 0x01      
        ## ReverseOnの設定値
        Reverse = 0x02
        



    #################################################################################################################

    # ////////////////////////////////////////////////
    # // ### 仮想の通信用の関数 ###
    # ////////////////////////////////////////////////
    
    def synchronize(self, txBuf, rxLen):
        """
        「synchronize」関数は、サブクラスでオーバーライドする必要があるプレースホルダーです。
        txBufのコマンドを送受信させます。
        
        @param txBuf 送信データ
        @param rxLen 受信データ数
        @return rxData 受信データ
        
        @attention rxDataはtimeout等でデータがない場合があります。
        @warning MemREAD等データを送受信する際は通信を確実に行うので上位で定義してください。
        """

        raise NotImplementedError("Serial側でオーバライドさせてください")



    def synchronizeVariableRead(self, txBuf):
        """
        「synchronizeVariableRead」関数は、サブクラスでオーバーライドする必要があるプレースホルダーです。
        受信データ数がわからない場合でも送受信したい場合に使用します。
        
        @param txBuf 送信データ
        @return rxData 受信データ
        
        @attention rxDataはtimeout等でデータがない場合があります。
        """

        raise NotImplementedError("Serial側でオーバライドさせてください")



    def synchronizeNoRead(self, txBuf):
        """
        「synchronizeNoRead」関数は、サブクラスでオーバーライドする必要があるプレースホルダーです。
        ブロードキャストIDなどデータが返ってこない時に使用します。
    
        @param txBuf 送信データ
        @return True:送信成功、False:送信失敗
        """

        raise NotImplementedError("Serial側でオーバライドさせてください")



    def setSerialParameters(self, baudrate: int, parity, timeout: int):
        """
        「setSerialParameters」関数は、サブクラスでオーバーライドする必要があります。
        通信速度、パリティ、タイムアウトでPC等上位側の通信設定を変更する際に使用します。
        
        @param baudrate 通信速度
        @param parity パリティ
        @param timeout タイムアウト
        @return なし
        """

        raise NotImplementedError("Serial側でオーバライドさせてください")



    #################################################################################################################

    # ////////////////////////////////////////////////
    # // ### コマンド関連 ###　（コマンドで共通化できる処理）
    # ////////////////////////////////////////////////

    @staticmethod
    def defaultMakeCmd(id: int, cmd: int, header=0xfe):
        
        """
        通常のPMXのコマンドを作る関数です。

        @param id PMXサーボモータのID番号
        @param cmd コマンドの値
        @param header デフォルトのヘッダー値 基本は0xfe
        
        @return 生成されたコマンドのバイトのリスト
        """
        
        # デフォルトのコマンド生成
        #        header, header, id, Length                      , cmd,opt
        txbuf = [header, header, id, PmxLib.MinimumLength.Receive, cmd, 0]
        
        # CRCを計算し代入する
        crc = CRC16Class.getCrc16(txbuf)
        txbuf.extend(crc.to_bytes(2, byteorder="little"))
        
        return txbuf



    @staticmethod
    def checkRecv(rxBuff, cmd: int, header=0xfe):
        
        """
        checkRecv関数は、受信パケットのヘッダー、パケット長、およびコマンドをチェックします。
        
        @details 受信データを解析し、下記エラーかどうかチェックを行います。
        @details
        * ヘッダのチェック
        * timeout(コマンドを何か受信したか)
        * 受信データの長さ
        * 受信したコマンドが一致するかどうか
        * CRC
        
        @param rxBuff 受信したCRCを含む全てのデータ(バイト配列)
        @param cmd 受信時にチェックするコマンド
        @param header ヘッダーがデフォルトと違う場合に指定する
        
        @return (Uint16) 通信の状態(ComError参照)
        """
        
        # sizeチェック(timeout)
        if (len(rxBuff) <= 0):
            return PmxLib.ComError.TimeOut

        # 最低限の長さのチェック(ここがないと他のチェックでエラーになる場合がある)
        if len(rxBuff) < PmxLib.MinimumLength.Receive:
            return PmxLib.ComError.ReceiveError
        
        # ヘッダのチェック
        if rxBuff[0] != header or rxBuff[1] != header:
            return PmxLib.ComError.ReceiveError

        # コマンドの長さチェック
        if rxBuff[PmxLib.BuffPter.Length] != len(rxBuff):
            return PmxLib.ComError.ReceiveError

        # コマンドチェック
        if rxBuff[PmxLib.BuffPter.CMD] != (cmd & 0x7f):
            return PmxLib.ComError.ReceiveError

        # CRCのチェック
        if CRC16Class.checkCrc16(rxBuff) == False:
            return PmxLib.ComError.CrcError

        # コマンドのチェック終了、問題なし
        return PmxLib.ComError.OK



    #################################################################################################################

    # ////////////////////////////////////////////////
    # // ### MemREAD ###
    # ////////////////////////////////////////////////

    def MemREAD(self, id: int, addr: int, size: int):
        """
        MemREADコマンドでPMXサーボモータからデータを取得します。
        
        @param id PMXサーボモータのID番号
        @param addr データを読み取る先頭のアドレス
        @param size 取得するデータサイズ
        
        @retval status, rxdata
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rxdata 指定されたアドレスから読み取ったデータ配列(bytes型)。通信失敗時にはNoneを返します。
        """

        PUT_MAX_DATA_LENGTH = 244   #MemREADで読み出せるデータの最大サイズ
        
        # データサイズを確認する
        if size == 0 or size >= PUT_MAX_DATA_LENGTH:
            return PmxLib.ComError.FormatError, []
        
        #
        # MemREADの送信コマンド作成
        #
        txbuf = [0xfe,  # Header
                 0xfe,  # Header
                 id,    # Servo ID
                 11,    # Length(11固定)
                 PmxLib.SendCmd.MemREAD,    # Command
                 0]     # Option
        txbuf.extend(addr.to_bytes(2, byteorder="little"))  # Address
        txbuf.append(size)     # data Size
        crc = CRC16Class.getCrc16(txbuf)  # CRCを計算
        txbuf.extend(crc.to_bytes(2, byteorder="little"))   # CRCを挿入
        
        ''' 返信データ数'''
        rxlen = PmxLib.MinimumLength.Receive + size  # 返信データ数はパケットの最小Byte数 + 読み出すデータByte数

        #
        # コマンドの送受信を行う
        #
        rxbuf = self.synchronize(txbuf, rxlen)   

        #
        # 受信パケットの解析
        #
        # 受信パケットのヘッダー、パケット長、およびコマンドをチェック
        errorFlag = self.checkRecv(rxbuf, PmxLib.SendCmd.MemREAD)
        if errorFlag != PmxLib.ComError.OK:
            return errorFlag, []    # データに異常があった場合はエラー情報と空データを返す
            
        # 受信パケットのステータス(エラー情報)を取り出す
        status = rxbuf[PmxLib.BuffPter.Status]

        # 受信データの代入
        rxdata = copy.copy(rxbuf[PmxLib.BuffPter.Data:-2])
        
        # ステータスとデータを返す
        return status, rxdata


            
    def MemREADToByte(self, id: int, addr: int):
        """
        MemREADで取得したデータをbyte(符号なし1byte)に変換して取得します。
        
        @param id PMXサーボモータのID番号
        @param addr データを読み取る先頭のアドレス
        
        @retval [status, rxData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rxData 指定されたアドレスから読み取ったbyte(符号なし1byte)のデータ。通信失敗時にはNoneを返します。
        """
        
        status, oneData = self.MemREAD(id, addr, 1)
        
        reData = None        

        # 正常なデータが返ってきたときの処理
        if len(oneData) == 1:
            reData = int.from_bytes(oneData, byteorder="little", signed=False)  #
        
        return status, reData


        
    def MemREADToInt16(self, id: int, addr: int):
        """
        MemREADで取得したデータをint16(符号あり2byte)に変換して取得します。
        
        @param id PMXサーボモータのID番号
        @param addr データを読み取る先頭のアドレス
        
        @retval [status, rxData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rxData 指定されたアドレスから読み取ったint16(符号あり2byte)のデータ。通信失敗時にはNoneを返します。
        """

        status, byteArray = self.MemREAD(id, addr, 2)
        
        reData = None        

        # 正常なデータが返ってきたときの処理
        if len(byteArray) == 2:
            reData = int.from_bytes(byteArray, byteorder="little", signed=True)  # int16 リトルエンディアン
            
        return status, reData



    def MemREADToUint16(self, id: int, addr: int):
        """
        MemREADで取得したデータをUint16(符号なし2byte)に変換して取得します。
        
        @param id PMXサーボモータのID番号
        @param addr データを読み取る先頭のアドレス
        
        @retval [status, rxData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rxData 指定されたアドレスから読み取ったUint16(符号なし2byte)のデータ。通信失敗時にはNoneを返します。
        """

        status, byteArray = self.MemREAD(id, addr, 2)

        reData = None        

        # 正常なデータが返ってきたときの処理
        if len(byteArray) == 2:
            reData = int.from_bytes(byteArray, byteorder="little", signed=False)  # uint16 リトルエンディアン
            
        return status, reData



    def MemREADToInt32(self, id: int, addr: int):
        """
        MemREADで取得したデータをint32(符号あり4byte)に変換して取得します。
        
        @param id PMXサーボモータのID番号
        @param addr データを読み取る先頭のアドレス

        @retval [status, rxData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rxData 指定されたアドレスから読み取ったint32(符号あり4byte)のデータ。通信失敗時にはNoneを返します。
        """

        status, byteArray = self.MemREAD(id, addr, 4)
        
        reData = None        

        # 正常なデータが返ってきたときの処理
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray, byteorder="little", signed=True)  # int32 リトルエンディアン
        
        return status, reData    



    def MemREADToUint32(self, id: int, addr: int):
        """
        MemREADで取得したデータをUint32(符号なし4byte)に変換して取得します。
        
        @param id PMXサーボモータのID番号
        @param addr データを読み取る先頭のアドレス

        @retval [status, rxData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rxData 指定されたアドレスから読み取ったUint32(符号なし4byte)のデータ。通信失敗時にはNoneを返します。
        """

        status, byteArray = self.MemREAD(id, addr, 4)

        reData = None

        # 正常なデータが返ってきたときの処理
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray, byteorder="little", signed=False)  # uInt32 リトルエンディアン
        
        return status, reData



    #################################################################################################################

    # ////////////////////////////////////////////////
    # // ### MemWRITE ###
    # ////////////////////////////////////////////////

    def MemWRITE(self, id: int, baseAddr: int, dataByteArray, writeOpt=0):
        """
        PMXサーボモータにMemWRITEコマンドを発行し、結果を取得します。
        
        @param id PMXサーボモータのID番号
        @param baseAddr データを書き込む先頭アドレス
        @param dataByteArray 書き込むデータのbyteリスト
        @param writeOpt MemWRITEで使用するオプション 0:通常書き込み、1:TorqueOn中の強制書き込み
        
        @return status MemWRITEの送受信結果(送信結果+PMXのstatus(Uint16))
        """
        
        PUT_MAX_DATA_LENGTH = 245   # MemWRITEで書き込めるデータの最大Byte数
        
        # データの最大値の判定
        if (len(dataByteArray) <= 0) or (len(dataByteArray) > PUT_MAX_DATA_LENGTH):
            return PmxLib.ComError.FormatError

        #
        # MemWRITEの送信コマンド作成
        #
        txbuf = [0xfe,  # Header
                 0xfe,  # Header
                 id,    # Servo ID
                 10 + len(dataByteArray),    # Length 送信データ数はパケットの最小Byte数 + 書き込むデータByte数
                 PmxLib.SendCmd.MemWRITE,    # Command
                 writeOpt]     # Option　(0:通常書き込み 1:TorqueOn中強制書き込み)
        txbuf.extend(baseAddr.to_bytes(2, byteorder="little"))  # Address
        txbuf.extend(dataByteArray)         # data
        crc = CRC16Class.getCrc16(txbuf)    # CRCを計算
        txbuf.extend(crc.to_bytes(2, byteorder="little"))   # CRCを挿入
        
        ''' 返信データ数'''
        rxlen = PmxLib.MinimumLength.Receive    # 返信データ数はパケットの最小Byte数
        
        #
        # コマンドの送受信を行う
        #
        rxbuf = self.synchronize(txbuf, rxlen)   
        
        #
        # 受信パケットの解析
        #
        #　受信パケットのヘッダー、パケット長、およびコマンドをチェック
        errorFlag = self.checkRecv(rxbuf, PmxLib.SendCmd.MemWRITE)
        if errorFlag != PmxLib.ComError.OK:
            return errorFlag    # データに異常があった場合はエラー情報を返す
            
        # 受信パケットのステータス(エラー情報)を取り出す
        status = rxbuf[PmxLib.BuffPter.Status]
        
        # ステータスを返す
        return status



    def MemWRITEToByte(self, id: int, baseAddr: int, data: int, writeOpt=0):
        """
        MemWRITEでbyte(符号なし1byte)のデータを書き込みます。
        その結果のステータス情報を返信します。
        
        @param id PMXサーボモータのID番号
        @param baseAddr データを書き込む先頭アドレス
        @param data byte(符号なし1byte)の書きこみデータ
        @param writeOpt MemWRITEで使用するオプション 0:通常書き込み、1:TorqueOn中の強制書き込み
        
        @return status MemWRITEの送受信結果(送信結果+PMXのstatus(Uint16))
        """
        
        # dataByteArray = data.to_bytes(1, byteorder="little", signed=False)
        dataByteArray = [data]
        status = self.MemWRITE(id, baseAddr, dataByteArray, writeOpt=writeOpt)

        return status



    def MemWRITEToInt16(self, id: int, baseAddr: int, data: int, writeOpt=0):
        """
        MemWRITEでint16(符号あり2byte)のデータを書き込みます。
        その結果のステータス情報を返信します。
        
        @param id PMXサーボモータのID番号
        @param baseAddr データを書き込む先頭アドレス
        @param data int16(符号あり2byte)の書きこみデータ
        @param writeOpt MemWRITEで使用するオプション 0:通常書き込み、1:TorqueOn中の強制書き込み
        
        @return status MemWRITEの送受信結果(送信結果+PMXのstatus(Uint16))
        """
        
        dataByteArray = data.to_bytes(2, byteorder="little", signed=True)
        status = self.MemWRITE(id, baseAddr, dataByteArray, writeOpt=writeOpt)

        return status



    def MemWRITEToUint16(self, id: int, baseAddr: int, data: int, writeOpt=0):
        """
        MemWRITEでint16(符号なし2byte)のデータを書き込みます。
        その結果のステータス情報を返信します。
        
        @param id PMXサーボモータのID番号
        @param baseAddr データを書き込む先頭アドレス
        @param data Uint16(符号なし2byte)の書きこみデータ
        @param writeOpt MemWRITEで使用するオプション 0:通常書き込み、1:TorqueOn中の強制書き込み
        
        @return status MemWRITEの送受信結果(送信結果+PMXのstatus(Uint16))
        """
        
        dataByteArray = data.to_bytes(2, byteorder="little", signed=False)
        status = self.MemWRITE(id, baseAddr, dataByteArray, writeOpt=writeOpt)

        return status



    def MemWRITEToInt32(self, id: int, baseAddr: int, data: int, writeOpt=0):
        """
        MemWRITEでint32(符号あり4byte)のデータを書き込みます。
        その結果のステータス情報を返信します。
        
        @param id PMXサーボモータのID番号
        @param baseAddr データを書き込む先頭アドレス
        @param data int32(符号あり4byte)の書きこみデータ
        @param writeOpt MemWRITEで使用するオプション 0:通常書き込み、1:TorqueOn中の強制書き込み
        
        @return status MemWRITEの送受信結果(送信結果+PMXのstatus(Uint16))
        """

        dataByteArray = data.to_bytes(4, byteorder="little", signed=True)
        status = self.MemWRITE(id, baseAddr, dataByteArray, writeOpt=writeOpt)

        return status



    def MemWRITEToUint32(self, id: int, baseAddr: int, data: int, writeOpt=0):
        """
        MemWRITEでUint32(符号なし4byte)のデータを書き込みます。
        その結果のステータス情報を返信します。
        
        @param id PMXサーボモータのID番号
        @param baseAddr データを書き込む先頭アドレス
        @param data Uint32(符号なし4byte)の書きこみデータ
        @param writeOpt MemWRITEで使用するオプション 0:通常書き込み、1:TorqueOn中の強制書き込み
        
        @return status MemWRITEの送受信結果(送信結果+PMXのstatus(Uint16))
        """
        
        dataByteArray = data.to_bytes(4, byteorder="little", signed=False)
        status = self.MemWRITE(id, baseAddr, dataByteArray, writeOpt=writeOpt)

        return status



    #################################################################################################################

    # ////////////////////////////////////////////////////////////////////
    # // LOADの送受信関数	
    # ////////////////////////////////////////////////////////////////////
    
    def LOAD(self, id: int):
        """
        PMXサーボモータのLOADコマンドを発行し、ステータス結果を取得します。
        
        @param id PMXサーボモータのID番号
        @return status LOADの送受信結果(送信結果+PMXのstatus(Uint16))
        """
        
        #
        # LOADの送信コマンド作成
        #
        txbuf = [0xfe,  # Header
                 0xfe,  # Header
                 id,    # Servo ID
                 PmxLib.MinimumLength.Send,  # Length
                 PmxLib.SendCmd.LOAD,        # Command
                 0]     # Option
        crc = CRC16Class.getCrc16(txbuf)     # CRCを計算
        txbuf.extend(crc.to_bytes(2, byteorder="little"))   # CRCを挿入
        
        ''' 返信データ数'''
        rxlen = PmxLib.MinimumLength.Receive    # 返信データ数はパケットの最小Byte数
        
        #
        # コマンドの送受信を行う
        #
        rxbuf = self.synchronize(txbuf, rxlen)   
        
        #
        # 受信パケットの解析
        #
        # 受信パケットのヘッダー、パケット長、およびコマンドをチェック
        errorFlag = self.checkRecv(rxbuf, PmxLib.SendCmd.LOAD)
        if errorFlag != PmxLib.ComError.OK:
            return errorFlag    # データに異常があった場合はエラー情報と空データを返す
            
        # 受信パケットのステータス(エラー情報)を取り出す
        status = rxbuf[PmxLib.BuffPter.Status]
        
        # ステータスを返す
        return status
        
        

    #################################################################################################################
    
    # ////////////////////////////////////////////////////////////////////
    # // SAVEの送受信関数	
    # ////////////////////////////////////////////////////////////////////

    def SAVE(self, id: int):
        """
        PMXサーボモータのSAVEコマンドを発行し、ステータス結果を取得します。
        
        @param id サーボモータのID番号
        @return status SAVEコマンドの送信結果(送信結果+PMXのstatus(Uint16))
        """

        #
        # SAVEの送信コマンド作成
        #
        txbuf = [0xfe,  # Header
                 0xfe,  # Header
                 id,    # Servo ID
                 PmxLib.MinimumLength.Send,  # Length
                 PmxLib.SendCmd.SAVE,        # Command
                 0]     # Option
        crc = CRC16Class.getCrc16(txbuf)     # CRCを計算
        txbuf.extend(crc.to_bytes(2, byteorder="little"))   # CRCを挿入
        
        ''' 返信データ数'''
        rxlen = PmxLib.MinimumLength.Receive    # 返信データ数はパケットの最小Byte数
        
        #
        # コマンドの送受信を行う
        #
        rxbuf = self.synchronize(txbuf, rxlen)   
        
        #
        # 受信パケットの解析
        #
        # 受信パケットのヘッダー、パケット長、およびコマンドをチェック
        errorFlag = self.checkRecv(rxbuf, PmxLib.SendCmd.SAVE)
        if errorFlag != PmxLib.ComError.OK:
            return errorFlag    # データに異常があった場合はエラー情報と空データを返す
            
        # 受信パケットのステータス(エラー情報)を取り出す
        status = rxbuf[PmxLib.BuffPter.Status]
        
        # ステータスを返す
        return status



    #################################################################################################################

    # ////////////////////////////////////////////////////////////////////
    # // MotorREADの送受信関数	
    # ////////////////////////////////////////////////////////////////////

    def MotorREAD(self, id: int, receiveMode: int, controlMode=0x01):
        """
        MotorREAD関数を発行し、応答モードに応じてデータを取得します。
        
        @param id サーボモータのID番号
        @param receiveMode 応答モード。事前に応答モードをMemWrite関数やsetMotorReceive関数で設定する必要があります。
        @param controlMode 制御モード。デフォルト値は `0x01` です。位置制御モードの時の現在位置を判定するときに使用します。
        
        @retval [status, torqueSwitch, rxDataList]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval torqueSwitch FreeやTorqueOnの値(TorqueSwitchTypeを参照)
        @retval rxDataList MotorReadで読み取ったデータ[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧](合計8項目)
                
        @attention  rxDataListでデータがない場合はリストの中身の各データがNoneになります。
        """

        #
        # MotorREADの送信コマンド作成
        #
        txbuf = [0xfe,  # Header
                 0xfe,  # Header
                 id,    # Servo ID
                 PmxLib.MinimumLength.Send,  # Length
                 PmxLib.SendCmd.MotorREAD,   # Command
                 0]     # Option
        crc = CRC16Class.getCrc16(txbuf)     # CRCを計算
        txbuf.extend(crc.to_bytes(2, byteorder="little"))   # CRCを挿入
        
        ''' 返信データ数'''
        rxlen = PmxLib.MinimumLength.Receive + 1 + self.__byteCounter(receiveMode)    # 返信データ数はパケットの最小Byte数 + 応答モードで指定した項目のByte数
        
        #
        # コマンドの送受信を行う
        #
        rxbuf = self.synchronize(txbuf, rxlen)   
        
        #
        # 受信パケットの解析
        #
        # 受信パケットのヘッダー、パケット長、およびコマンドをチェック
        errorFlag = self.checkRecv(rxbuf, PmxLib.SendCmd.MotorREAD)
        if errorFlag != PmxLib.ComError.OK:
            return errorFlag, 0xff, []    # データに異常があった場合はエラー情報と空データを返す
            
        # 受信パケットのステータス(エラー情報)を取り出す
        status = rxbuf[PmxLib.BuffPter.Status]
        
        # トルクスイッチのデータを代入
        torqueSwitch = rxbuf[PmxLib.BuffPter.Data]
        
        # もし応答モードのデータがあったら読み出したデータを代入
        if (len(rxbuf) > 9):
            
            reDataBytes = copy.copy(rxbuf[PmxLib.BuffPter.Data + 1:-2])   # データ部分の抜き出し
            rxDataList = self.__convReceiveMotorData(receiveMode, reDataBytes, controlMode=controlMode)   # データをリストに変換
            
        else:
            rxDataList = []
        
        # ステータスを返す
        return status, torqueSwitch, rxDataList



    def __convReceiveMotorData(self, receiveMode: int, returnDataBytes, controlMode=0x01):
        """
        受信モードに基づいてMotorWrite/Readで送られて来たデータを変換します。
        
        @param receiveMode 受信モード設定
        @param returnDataBytes 受信されたデータのbyte配列(コマンド等は含まない)
        @param controlMode 制御モード(位置制御のみ符号ありで変換するために使用する) 
        @return 受信データを変換したモータデータを返す[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧]
        @attention データがない場合はリストの中身の各データがNoneになります
        """
        
        # dataListの定義[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧]
        reDataList = [None, None, None, None, None, None, None, None]

        # データbyte数と変換するデータ数が違っていたら空リストを返す
        if (len(returnDataBytes) != (self.__byteCounter(receiveMode))):
            return reDataList 
    
        arrayCount = 0  # データを数える(bytes配列の場所を決めるのに使用)
        
        #
        # 位置データ
        #
        if ((receiveMode & PmxLib.ReceiveDataOption.Position) != 0):
            
            psbuf = copy.copy(returnDataBytes[arrayCount * 2: arrayCount * 2 + 2])  # 2byte分コピー
            # 位置制御モードが入っている時は符号あり
            if (controlMode & PmxLib.ControlMode.Position) == PmxLib.ControlMode.Position:
                posData = int.from_bytes(psbuf, byteorder='little', signed=True)
            else:
                posData = int.from_bytes(psbuf, byteorder='little', signed=False)
            arrayCount += 1
            # reDataList.append(posData)
            reDataList[0] = posData
            
        #
        # 速度データ
        #
        if ((receiveMode & PmxLib.ReceiveDataOption.Speed) != 0):
            sbuf = copy.copy(returnDataBytes[arrayCount * 2: arrayCount * 2 + 2])  # 2byte分コピー
            spdData = int.from_bytes(sbuf, byteorder='little', signed=True)
            arrayCount += 1
            # reDataList.append(spdData)
            reDataList[1] = spdData
    
        #
        # 電流データ
        #
        if ((receiveMode & PmxLib.ReceiveDataOption.Current) != 0):
            cbuf = copy.copy(returnDataBytes[arrayCount * 2: arrayCount * 2 + 2])  # 2byte分コピー
            curData = int.from_bytes(cbuf, byteorder='little', signed=True)
            arrayCount += 1
            # reDataList.append(curData)
            reDataList[2] = curData

        #
        # トルクデータ
        #
        if ((receiveMode & PmxLib.ReceiveDataOption.Torque) != 0):
            tbuf = copy.copy(returnDataBytes[arrayCount * 2: arrayCount * 2 + 2])  # 2byte分コピー
            trqData = int.from_bytes(tbuf, byteorder='little', signed=True)
            arrayCount += 1
            # reDataList.append(trqData)
            reDataList[3] = trqData

        #
        # PWM出力データ
        #
        if ((receiveMode & PmxLib.ReceiveDataOption.Pwm) != 0):
            pbuf = copy.copy(returnDataBytes[arrayCount * 2: arrayCount * 2 + 2])  # 2byte分コピー
            pwmData = int.from_bytes(pbuf, byteorder='little', signed=True)
            arrayCount += 1
            # reDataList.append(pwmData)
            reDataList[4] = pwmData

        #
        # モータ温度データ
        #
        if ((receiveMode & PmxLib.ReceiveDataOption.MotorTemp) != 0):
            mbuf = copy.copy(returnDataBytes[arrayCount * 2: arrayCount * 2 + 2])  # 2byte分コピー
            mTempData = int.from_bytes(mbuf, byteorder='little', signed=True)
            arrayCount += 1
            # reDataList.append(mTempData)
            reDataList[5] = mTempData

        #
        # CPU温度データ
        #
        if ((receiveMode & PmxLib.ReceiveDataOption.CpuTemp) != 0):
            ctbuf = copy.copy(returnDataBytes[arrayCount * 2: arrayCount * 2 + 2])  # 2byte分コピー
            cTempData = int.from_bytes(ctbuf, byteorder='little', signed=True)
            arrayCount += 1
            # reDataList.append(cTempData)
            reDataList[6] = cTempData

        #
        # 電圧データ
        #
        if ((receiveMode & PmxLib.ReceiveDataOption.Voltage) != 0):
            vbuf = copy.copy(returnDataBytes[arrayCount * 2: arrayCount * 2 + 2])  # 2byte分コピー
            volData = int.from_bytes(vbuf, byteorder='little', signed=False)
            arrayCount += 1
            # reDataList.append(volData)
            reDataList[7] = volData
        
        return reDataList
    


    # 制御モード、応答モードからByte数をカウントする
    def __byteCounter(self, val: int):
        """
        制御モード、応答モードを基に必要とするデータが合計何byteになるのか計算をします。
        
        @param val 応答モードもしくは制御モード
        
        @return 制御モードや応答モードに必要とするデータ(byte)数
        """
        
        cnt = 0
        
        # bitが立っている個数の計算
        for i in range(8):
            check = val & 0x01
            if check == 1:
                cnt += 1
            
            val = val >> 1
        
        # 1data2byteなので、2倍の数値を返す
        return cnt * 2



    #################################################################################################################

    # ////////////////////////////////////////////////////////////////////
    # // MotorWRITEの送受信関数	
    # ////////////////////////////////////////////////////////////////////
    
    def MotorWRITE(self, id: int, mototWRITEopt: int, controlMode=0x01, receiveMode=0, sendDataList=[]):
        """
        MotorWRITE 機能は、モータの制御やトルクのオン/オフを切り替えるコマンドを送信し、モーターからステータスとデータを受信します。
        
        @param id PMXサーボモータのID番号
        @param mototWRITEopt コントロールモードのデータを送るかFree/Holdにするかを決めるのに使用します
        @param controlMode モータの制御モード。応答モードで現在位置を取得する時のみに使用します
        @param receiveMode 応答モード。返ってきたデータを変換するときに使用します。
        @param sendDataList モータへの指示値を格納
    
        @retval [status, torqueSwitch, rxDataList]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval torqueSwitch FreeやTorqueOnの値(TorqueSwitchTypeを参照)
        @retval rxDataList MotorReadで読み取ったデータ[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧](合計8項目)
        
        @attention sendDataListは、位置 > 速度 > 電流 > トルク > PWM > 時間 の順番に格納される
        @attention 位置/電流制御モードの場合は[position,current]をsendDataListに渡します
        @attention sendDataListは使わないデータ(位置/電流の時のPWM等)は書かない
        
        @warning sendDataListの順番を間違えない事
        @warning 複数組み合わせの場合で１つだけ変えるときでも全ての目標値を送ってください。
        @warning controlModeやreceiveModeは事前に設定をする事
        
        @attention rxDataListでデータがない場合はリストの中身の各データがNoneになります
        """
        
        # トルクスイッチの場合の送信パケットサイズ
        if (mototWRITEopt == PmxLib.TorqueSwitchType.TorqueOn or 
            mototWRITEopt == PmxLib.TorqueSwitchType.Free or 
            mototWRITEopt == PmxLib.TorqueSwitchType.Brake or
            mototWRITEopt == PmxLib.TorqueSwitchType.Hold):
            
            txlen = PmxLib.MinimumLength.Send     # 送信パケットのバイト数       
            
        # サーボを制御する場合の送信パケットサイズ
        elif mototWRITEopt == PmxLib.TorqueSwitchType.Control:
            if (controlMode == 0):  # controlModeが0でmototWRITEoptが0の場合はありえないのでエラーを返す
                return PmxLib.ComError.FormatError, 0xff, []
            
            txlen = PmxLib.MinimumLength.Send + self.__byteCounter(controlMode)     # 送信パケットのバイト数
            
        else:
            return PmxLib.ComError.FormatError, 0xff, []

        #
        # MotorWRITEの送信コマンド作成
        #
        txbuf = [0xfe,   # Header
                 0xfe,   # Header
                 id,     # Servo ID
                 txlen,  # Length
                 PmxLib.SendCmd.MotorWRITE,       # Command
                 mototWRITEopt]     # Option


        # 制御モードの組み合わせによる分岐
        if (mototWRITEopt == 0x00) and (len(sendDataList) > 0):
            
            # if txlen == 10:
            #     txbuf.extend(data1.to_bytes(2, byteorder="little"))   # 指令値1を挿入
                
            # elif txlen == 12:
            #     txbuf.extend(data1.to_bytes(2, byteorder="little"))   # 指令値1を挿入
            #     txbuf.extend(data2.to_bytes(2, byteorder="little"))   # 指令値2を挿入

            # elif txlen == 14:
            #     txbuf.extend(data1.to_bytes(2, byteorder="little"))   # 指令値1を挿入
            #     txbuf.extend(data2.to_bytes(2, byteorder="little"))   # 指令値2を挿入
            #     txbuf.extend(data3.to_bytes(2, byteorder="little"))   # 指令値3を挿入
            
            
            for sendData in sendDataList:
                # データがsigned/Unsignedごちゃまぜで来るので処理
                addData = 0
                if (sendData < 0):
                    addData = sendData.to_bytes(2, byteorder="little", signed=True) 
                else:
                    addData = sendData.to_bytes(2, byteorder="little", signed=False)
                
                txbuf.extend(addData)

            
        # オプションがトルクON等々の場合はデータなし
        # elif opt == 0x01 or opt == 0x02 or opt == 0x04 or opt == 0x08:
        elif (mototWRITEopt == self.TorqueSwitchType.TorqueOn or 
              mototWRITEopt == self.TorqueSwitchType.Free or 
              mototWRITEopt == self.TorqueSwitchType.Brake or 
              mototWRITEopt == self.TorqueSwitchType.Hold):
            pass
        
        else:  # オプションがおかしい場合
            return PmxLib.ComError.FormatError, self.TorqueSwitchType.Error, []

        crc = CRC16Class.getCrc16(txbuf)    # CRCを計算
        txbuf.extend(crc.to_bytes(2, byteorder="little"))   # CRCを挿入

        ''' 返信データ数'''
        rxlen = PmxLib.MinimumLength.Receive + 1 + self.__byteCounter(receiveMode)    # 返信データ数はパケットの最小Byte数 + 応答モードで指定した項目のByte数
        
        #
        # コマンドの送受信を行う
        #
        rxbuf = self.synchronize(txbuf, rxlen)   
        
        #
        # 受信パケットの解析
        #
        # 受信パケットのヘッダー、パケット長、およびコマンドをチェック
        errorFlag = self.checkRecv(rxbuf, PmxLib.SendCmd.MotorWRITE)
        if errorFlag != PmxLib.ComError.OK:
            return errorFlag, 0xff, []  # データに異常があった場合はエラー情報と空データを返す
            
        # 受信パケットのステータス(エラー情報)を取り出す
        status = rxbuf[PmxLib.BuffPter.Status]
        
        # トルクスイッチのデータを代入
        torqueSwitch = rxbuf[PmxLib.BuffPter.Data]
        
        # もし応答モードのデータがあったら読み出したデータを代入
        rxdataList = []
        if (len(rxbuf) > 9):
            rxBytesDatas = copy.copy(rxbuf[PmxLib.BuffPter.Data + 1:-2])
            rxdataList = self.__convReceiveMotorData(receiveMode, rxBytesDatas, controlMode=controlMode)
            
        else:
            rxdataList = []
        
        # ステータスを返す
        return status, torqueSwitch, rxdataList



    def MotorWRITESingle(self, id: int, controlMode: int, receiveMode: int, targetVal: int):
        """
        MotorWRITEでデータ指示一つの場合に使用します
        
        @param id PMXサーボモータのID番号
        @param controlMode モータの制御モード。応答モードで現在位置を取得する時のみに使用します。
        @param receiveMode 応答モード。返ってきたデータを変換するときに使用します。
        @param targetVal 目標値(位置、速度、電流、トルク、PWM)が入ります
                
        @retval [status, torqueSwitch, rxDataList]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval torqueSwitch FreeやTorqueOnの値(TorqueSwitchTypeを参照)
        @retval rxDataList MotorReadで読み取ったデータ[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧](合計8項目)
        
        @warning controlModeやreceiveModeは事前に設定をする事
        
        @attention rxDataListでデータがない場合はリストの中身の各データがNoneになります
        """
        
        return self.MotorWRITE(id, 0, controlMode=controlMode, receiveMode=receiveMode, sendDataList=[targetVal])
    
    
        
    def MotorWRITEDouble(self, id: int, controlMode: int, receiveMode: int, targetVal1: int, targetVal2: int):
        """
        MotorWriteを使用して2変数を指示するときに使用します。
        
        @param id PMXサーボモータのID番号
        @param controlMode モータの制御モード。応答モードで現在位置を取得する時のみに使用します。
        @param receiveMode 応答モード。返ってきたデータを変換するときに使用します。
        @param targetVal1 目標値1(位置、速度、電流、トルク、PWM)が入ります
        @param targetVal2 目標値2(速度、電流、トルク、PWM、時間)が入ります

        @retval [status, torqueSwitch, rxDataList]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval torqueSwitch FreeやTorqueOnの値(TorqueSwitchTypeを参照)
        @retval rxDataList MotorReadで読み取ったデータ[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧](合計8項目)
        
        @attention 位置/電流制御モードの場合は
        @attention targetVal1 = position,targetVal2 = current
        @attention を渡します
        
        @warning targetValの順番を間違えない事
        @warning controlModeやreceiveModeは事前に設定をする事
        
        @attention rxDataListでデータがない場合はリストの中身の各データがNoneになります
        """
        
        return self.MotorWRITE(id, 0, controlMode=controlMode, receiveMode=receiveMode, sendDataList=[targetVal1, targetVal2])



    def MotorWRITETriple(self, id: int, controlMode: int, receiveMode: int, targetVal1: int, targetVal2: int, targetVal3: int):
        """
        MotorWriteを使用して3変数を指示するときに使用します。
        
        @param id PMXサーボモータのID番号
        @param controlMode モータの制御モード。応答モードで現在位置を取得する時のみに使用します。
        @param receiveMode 応答モード。返ってきたデータを変換するときに使用します。
        @param targetVal1 目標値1(位置、速度、電流、トルク、PWM)のどれかが入ります
        @param targetVal2 目標値2(速度、電流、トルク、PWM、時間)のどれかが入ります
        @param targetVal3 目標値3(電流、トルク、PWM、時間)のどれかが入ります

        @retval [status, torqueSwitch, rxDataList]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval torqueSwitch FreeやTorqueOnの値(TorqueSwitchTypeを参照)
        @retval rxDataList MotorReadで読み取ったデータ[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧](合計8項目)
        
        @attention 位置/トルク/時間制御モードの場合は
        @attention targetVal1 = position, targetVal2 = torque, targetVal3 = time
        @attention を渡します
        @warning targetValの順番を間違えない事
        @warning controlModeやreceiveModeは事前に設定をする事
        
        @attention rxDataListでデータがない場合はリストの中身の各データがNoneになります
        """

        return self.MotorWRITE(id, 0, controlMode=controlMode, receiveMode=receiveMode, sendDataList=[targetVal1, targetVal2, targetVal3])
    


    #################################################################################################################

    # ////////////////////////////////////////////////////////////////////
    # // SystemREADの送受信関数	
    # ////////////////////////////////////////////////////////////////////
    
    def SystemREAD(self, id: int):
        """
        SystemREAD 関数は、サーボからシステム情報を読み取ります。
        
        @details 読み取ることができるデータは下記になります。
        @details * PMX個体固有のシリアル番号
        @details * PMXの機種番号
        @details * バージョンの情報
        @details * コマンドを受け取ってから返信するまでの時間
        
        @param id PMXサーボモータのID番号
        @retval [status, returnData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval returnData SystemREADで取得したすべてのデータを返します。[SerialNum, ModelNum, Version, responsTime]
        
        @attention Version => [MAJOR,MINOR,PATCH,BUILD]
        @attention Version以外は数値が返ります
        """
        
        #
        # SystemREADの送信コマンド作成
        #
        txbuf = [0xfe,  # Header
                 0xfe,  # Header
                 id,    # Servo ID
                 PmxLib.MinimumLength.Send,  # Length
                 PmxLib.SendCmd.SystemREAD,  # Command
                 0]     # Option
        crc = CRC16Class.getCrc16(txbuf)    # CRCを計算
        txbuf.extend(crc.to_bytes(2, byteorder="little"))   # CRCを挿入
        
        ''' 返信データ数'''
        rxlen = 21
        
        # 返信データは作っておく(SerialNum,ModelName,F/W,returnTime)
        returnData = [None, None, None, None]
        
        #
        # コマンドの送受信を行う
        #
        rxbuf = self.synchronize(txbuf, rxlen)   
        
        #
        # 受信パケットの解析
        #
        # 受信パケットのヘッダー、パケット長、およびコマンドをチェック
        errorFlag = self.checkRecv(rxbuf, PmxLib.SendCmd.SystemREAD)
        if errorFlag != PmxLib.ComError.OK:
            return errorFlag, returnData  # データに異常があった場合はエラー情報を返す
            
        # 受信パケットのステータス(エラー情報)を取り出す
        status = rxbuf[PmxLib.BuffPter.Status]
        
        # コマンド、RAMアクセス、Mode,dataエラーの時は正常値が返ってこないのでエラー値を返す
        if (status & 0x7C) != 0x00:
            return status, returnData

        
        # データ部分を抜き出し
        rxdata = copy.copy(rxbuf[PmxLib.BuffPter.Data:-2])
        
        returnData[0] = int.from_bytes(rxdata[0:4], byteorder="little", signed=False)  # SerialNum  
        returnData[1] = int.from_bytes(rxdata[4:8], byteorder="little", signed=False)  # ModelNum  
        returnData[2] = [rxdata[11], rxdata[10], rxdata[9], rxdata[8]]  # MAJOR,MINOR,PATCH,BUILD
        returnData[3] = rxdata[12]  # returnTime
        
        # ステータスを返す
        return status, returnData
        
        


    #################################################################################################################
        
    # ////////////////////////////////////////////////////////////////////
    # // SystemWRITEの送受信関数	
    # ////////////////////////////////////////////////////////////////////
        
    def SystemWRITE(self, id: int, serialNum=None, newId=None, newBaudrateVal=None, newParityVal=None, newResponseTime=None):
        """
        SystemWRITEコマンドを使用してID、ボーレート、パリティ、および応答時間を設定します。
        
        @param id PMXサーボモータのID番号
        @param serialNum サーボの固有番号。SystemReadでデータを返します
        @param newId サーボに割り当てる新しい ID。 0 ～ 239 の整数値である必要があります。
        @param newBaudrateVal 新しい通信速度設定値。直接書き込む値なので、EditBaudrateの値を使用します
        @param newParityVal パリティを変更します。直接値を書き込むので、EditParityの値を使用します
        @param newResponseTime コマンドを受け取ってから返信するまでの時間を設定します(us単位で設定します)
        @return status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        
        @note パラメータは個別に書き変えることができます。
        @note パラメータを渡さないもしくはNoneにすると変更されません
        
        @attention SystemWrite使用時にはシリアル番号がNoneの場合は関数内で読み込みます。
        """

        # シリアル番号がなかったら取得する
        if serialNum is None: 
            # serial番号を取得する
            status, serialNum = self.getSerialNumber(id)
            
            if serialNum is None:  # シリアル番号取得に失敗したらステータスを返す 
                return status

        sysWriteOpt = 0x00  # optionは途中で計算する

        # 
        # ID
        #         
        PUT_MAX_ID_DATA = 239   # IDの最大値
        
        if newId is None: 
            newId = 0xFF  # なんでもいいがありえない値を入れておく
        elif newId < 0 or newId > PUT_MAX_ID_DATA:
            return PmxLib.ComError.FormatError
        else:
            sysWriteOpt = sysWriteOpt | 0x01
        
        #
        # Baudrate
        #
        PUT_MAX_BAUDRATE_DATA = 7   # MemWRITEで書き込めるデータの最大Byte数

        if newBaudrateVal is None:
            newBaudrateVal = 0xFF  # なんでもいいがありえない値を入れておく
        # 書き換える通信速度が範囲内かを確認する
        elif newBaudrateVal < 0 or newBaudrateVal > PUT_MAX_BAUDRATE_DATA:
            return PmxLib.ComError.FormatError
        else:
            sysWriteOpt = sysWriteOpt | 0x02

        #
        # Parity
        #
        PUT_MAX_PARITY_DATA = 2   # MemWRITEで書き込めるデータの最大Byte数

        if newParityVal is None:
            newParityVal = 0xFF  # なんでもいいがありえない値を入れておく
        # 書き換えるパリティが範囲内かを確認する
        elif newParityVal < 0 or newParityVal > PUT_MAX_PARITY_DATA:
            return PmxLib.ComError.FormatError
        else:
            sysWriteOpt = sysWriteOpt | 0x04

        #
        # ResponseTime
        #
        
        if newResponseTime is None:
            newResponseTime = 0xFF
        else:
            sysWriteOpt = sysWriteOpt | 0x08
        
        #
        # SystemWRITEの送信コマンド作成
        #
        txbuf = [0xfe,  # Header
                 0xfe,  # Header
                 id,    # Servo ID
                 PmxLib.MinimumLength.Send + 8,  # Length
                 PmxLib.SendCmd.SystemWRITE,     # Command
                 sysWriteOpt]     # Option オプションにより書き換えるデータを指定　Bit0:IDデータの変更, Bit1: ボーレートの変更, Bit2: パリティの変更, Bit3:応答時間の変更
        txbuf.extend(serialNum.to_bytes(4, byteorder="little"))   # シリアル番号　※必ずSystemREADでサーボ固有のシリアル番号を読み出してからSystemWRITEを実行
        txbuf.append(newId)                 # 書き換えるID番号
        txbuf.append(newBaudrateVal)           # 書き換える通信速度
        txbuf.append(newParityVal)             # 書き換えるパリティ
        txbuf.append(newResponseTime)       # 書き換える応答時間
        crc = CRC16Class.getCrc16(txbuf)    # CRCを計算
        txbuf.extend(crc.to_bytes(2, byteorder="little"))   # CRCを挿入
        
        ''' 返信データ数'''
        rxlen = PmxLib.MinimumLength.Receive
        
        #
        # コマンドの送受信を行う
        #
        rxbuf = self.synchronize(txbuf, rxlen)   
        
        #
        # 受信パケットの解析
        #
        # 受信パケットのヘッダー、パケット長、およびコマンドをチェック
        errorFlag = self.checkRecv(rxbuf, PmxLib.SendCmd.SystemWRITE)
        if errorFlag != PmxLib.ComError.OK:
            return errorFlag    # データに異常があった場合はエラー情報を返す
            
        # 受信パケットのステータス(エラー情報)を取り出す
        status = rxbuf[PmxLib.BuffPter.Status]
        
        # ステータスを返す
        return status
        


    #################################################################################################################

    # ////////////////////////////////////////////////////////////////////
    # // ReBootCmdの送受信関数	
    # ////////////////////////////////////////////////////////////////////
    
    def ReBoot(self, id: int, resetTime=0):
        """
        サーボモータを再起動させます。
        
        @param id PMXサーボモータのID番号
        @param resetTime 再起動まで待つ時間(ms)
        
        @return status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        
        @attention ReBoot中はWrite等を行ってもエラーが返ってきます。
        @attention ReBoot直後は内部パラメータを読み込むまでデータは返ってきません
        
        """
        
        #
        # ReBootの送信コマンド作成
        #
        txbuf = [0xfe,  # Header
                 0xfe,  # Header
                 id,    # Servo ID
                 PmxLib.MinimumLength.Send + 2,  # Length
                 PmxLib.SendCmd.ReBoot,  # Command
                 0]     # Option
        txbuf.extend(resetTime.to_bytes(2, byteorder="little", signed=False))   # リセットを開始するまでの時間
        crc = CRC16Class.getCrc16(txbuf)    # CRCを計算
        txbuf.extend(crc.to_bytes(2, byteorder="little"))   # CRCを挿入
        
        ''' 返信データ数'''
        rxlen = PmxLib.MinimumLength.Receive
        
        #
        # コマンドの送受信を行う
        #
        rxbuf = self.synchronize(txbuf, rxlen)   
        
        #
        # 受信パケットの解析
        #
        # 受信パケットのヘッダー、パケット長、およびコマンドをチェック
        errorFlag = self.checkRecv(rxbuf, PmxLib.SendCmd.ReBoot)
        if errorFlag != PmxLib.ComError.OK:
            return errorFlag    # データに異常があった場合はエラー情報を返す
            
        # 受信パケットのステータス(エラー情報)を取り出す
        status = rxbuf[PmxLib.BuffPter.Status]
        
        # ステータスを返す
        return status
        
        


    #################################################################################################################

    # ////////////////////////////////////////////////////////////////////
    # // FactoryResetの送受信関数	
    # ////////////////////////////////////////////////////////////////////
    
    def FactoryReset(self, id: int, serialNum=None):
        """
        サーボを工場出荷時の設定にリセットします
        
        @param id PMXサーボモータのID番号
        @param serialNum シリアル番号
        @return status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        
        @note serialNumがNone場合はこの関数内で取得するので省略可能
        """

        # シリアル番号がなかったら取得する
        if serialNum is None: 
            # serial番号を取得する
            status, serialNum = self.getSerialNumber(id)
            
            if serialNum is None:  # シリアル番号取得に失敗したらステータスを返す 
                return status
        
        #
        # MemREADの送信コマンド作成
        #
        txbuf = [0xfe,  # Header
                 0xfe,  # Header
                 id,    # Servo ID
                 PmxLib.MinimumLength.Send + 4, # Length
                 PmxLib.SendCmd.FactoryReset, # Command
                 0]     # Option
        txbuf.extend(serialNum.to_bytes(4, byteorder="little"))   # リセットを開始するまでの時間
        crc = CRC16Class.getCrc16(txbuf)    # CRCを計算
        txbuf.extend(crc.to_bytes(2, byteorder="little"))   # CRCを挿入
        
        ''' 返信データ数'''
        rxlen = PmxLib.MinimumLength.Receive
        
        #
        # コマンドの送受信を行う
        #
        rxbuf = self.synchronize(txbuf, rxlen)   
        
        #
        # 受信パケットの解析
        #
        #　受信パケットのヘッダー、パケット長、およびコマンドをチェック
        errorFlag = self.checkRecv(rxbuf, PmxLib.SendCmd.FactoryReset)
        if errorFlag != PmxLib.ComError.OK:
            return errorFlag    #　データに異常があった場合はエラー情報を返す
            
        # 受信パケットのステータス(エラー情報)を取り出す
        status = rxbuf[PmxLib.BuffPter.Status]
        
        # ステータスを返す
        return status
    

        

    #################################################################################################################    
    
    #################################################################################################################
    
    # MemREADを利用したデータのget関数一覧
    
    #################################################################################################################


    def getPosition(self, id: int, contolMode=0x01):
        """
        指定したIDの現在の位置を取得します。
        
        @details MemREADを使用して指定したIDから現在位置を取得します。        
        @details PMXでは、位置制御モードが含まれるとデータが符号あり(±320deg)、含まれない場合は符号なし(0-360deg)のデータが返ってきます
        
        @param id PMXサーボモータのID番号
        @param contolMode 制御モードを入れます。デフォルトは0x01です。位置制御モードが含まれるかどうかに使用します。

        @retval [status, posData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval posData 現在の位置データ。位置制御モードによって符号有り無しが変わります 。
        """

        # 位置制御が含まれるときはSignedの値を返す
        if (contolMode & PmxLib.ControlMode.Position) == PmxLib.ControlMode.Position:
            status, posData = self.MemREADToInt16(id, PmxLib.RamAddrList.NowPosition)
            return status, posData
        # 位置制御が含まれない時はUnsignedで値を返す
        else:
            status, posData = self.MemREADToUint16(id, PmxLib.RamAddrList.NowPosition)
            return status, posData    



    def getSpeed(self, id: int):
        """
        指定した ID の現在の速度を取得します。

        @details MemREADを使用して指定したIDから現在速度を取得します。   
        
        @param id PMXサーボモータのID番号
        
        @retval [status, spdData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval spdData 現在の速度データ
        """

        status, spdData = self.MemREADToInt16(id, PmxLib.RamAddrList.NowSpeed)

        return status, spdData



    def getCurrent(self, id: int):
        """
        指定した ID の現在の電流を取得します。

        @details MemREADを使用して指定したIDから現在電流を取得します。   
        
        @param id PMXサーボモータのID番号
        
        @retval [status, curData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval curData 現在の電流データ
        """

        status, curData = self.MemREADToInt16(id, PmxLib.RamAddrList.NowCurrent)

        return status, curData



    def getTorque(self, id: int):
        """
        指定した ID の現在のトルクを取得します。

        @details MemREADを使用して指定したIDから現在の推定トルクを取得します。   

        @param id PMXサーボモータのID番号
        
        @retval [status, trqData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval trqData 現在の推定トルクデータ
        """

        status, trqData = self.MemREADToInt16(id, PmxLib.RamAddrList.NowTorque)

        return status, trqData



    def getPwm(self, id: int):
        """
        指定した ID の現在のPWMを取得します。

        @details MemREADを使用して指定したIDから現在PWMを取得します。   
        
        @param id PMXサーボモータのID番号
        
        @retval [status, pwmData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval pwmData 現在のPWMデータ
        """

        status, pwmData = self.MemREADToInt16(id, PmxLib.RamAddrList.NowPwm)

        return status, pwmData



    def getMotorTemp(self, id: int):
        """
        指定した ID の現在のモータ温度を取得します。

        @details MemREADを使用して指定したIDから現在モータ温度を取得します。   
        
        @param id PMXサーボモータのID番号
        
        @retval [status, motTempData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval motTempData 現在のモータ温度データ
        """

        status, motTempData = self.MemREADToInt16(id, PmxLib.RamAddrList.MotorTemp)

        return status, motTempData



    def getCPUTemp(self, id: int):
        """
        指定した ID の現在のCPU温度を取得します。

        @details MemREADを使用して指定したIDから現在CPU温度を取得します。   
        
        @param id PMXサーボモータのID番号
        
        @retval [status, cpuTempData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval cpuTempData 現在のCPU温度データ
        """

        status, cpuTempData = self.MemREADToInt16(id, PmxLib.RamAddrList.CPUTemp)

        return status, cpuTempData



    def getTrajectoryTime(self, id: int):
        """
        指定した ID の現在の補間時間を取得します。

        @details MemREADを使用して指定したIDから現在補間時間を取得します。   
        
        @param id PMXサーボモータのID番号
        
        @retval [status, traTimeData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval traTimeData 現在の補間時間データ
        """

        status, traTimeData = self.MemREADToUint16(id, PmxLib.RamAddrList.TrajectoryTime)

        return status, traTimeData



    def getInputVoltage(self, id: int):
        """
        指定した ID の現在の電圧を取得します。

        @details MemREADを使用して指定したIDから現在電圧を取得します。   
        
        @param id PMXサーボモータのID番号
        
        @retval [status, volData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval volData 現在の電圧データ
        """

        status, volData = self.MemREADToInt16(id, PmxLib.RamAddrList.InputVoltage)

        return status, volData



    def getEncoder(self, id: int):
        """
        指定した ID のエンコーダを取得します。

        @details MemREADを使用して指定したIDからエンコーダを取得します。   
        
        @param id PMXサーボモータのID番号
        
        @retval [status, encData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval encData エンコーダデータ
        """

        status, encData = self.MemREADToUint16(id, PmxLib.RamAddrList.EncoderValue)

        return status, encData    



    def getStatus(self, id: int):
        """
        指定した ID のステータスを取得します。

        @details MemREADを使用して指定したIDからステータスを取得します。
        @details 詳しくは取扱説明書をご覧ください。<A HREF= https://kondo-robot.com/faq/pmx-servo-series-online-manual>PMXのオンラインマニュアル</A>
        
        @param id PMXサーボモータのID番号
        
        @retval [status, stData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval stData ステータスデータ

        @attention この関数を呼ぶとサーボ内のエラー値はリセットされます。
        """

        status, stData = self.MemREADToByte(id, PmxLib.RamAddrList.ErrorStatus)

        return status, stData    



    def getSystemStatus(self, id: int):
        """
        指定した ID のシステムステータスを取得します。

        @details MemREADを使用して指定したIDからシステムステータスを取得します。
        @details 詳しくは取扱説明書をご覧ください。<A HREF= https://kondo-robot.com/faq/pmx-servo-series-online-manual>PMXのオンラインマニュアル</A>   
        
        @param id PMXサーボモータのID番号
        
        @retval [status, sysStaData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval sysStaData システムステータスデータ

        @attention この関数を呼ぶとサーボ内のエラー値はリセットされます。
        """

        status, sysStaData = self.MemREADToByte(id, PmxLib.RamAddrList.ErrorSystem)

        return status, sysStaData  



    def getMotorStatus(self, id: int):
        """
        指定した ID のモータステータスを取得します。

        @details MemREADを使用して指定したIDからモータステータスを取得します。
        @details 詳しくは取扱説明書をご覧ください。<A HREF= https://kondo-robot.com/faq/pmx-servo-series-online-manual>PMXのオンラインマニュアル</A>
        
        @param id PMXサーボモータのID番号
        
        @retval [status, motStaData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval motStaData モータステータスデータ

        @attention この関数を呼ぶとサーボ内のエラー値はリセットされます。
        """

        status, motStaData = self.MemREADToByte(id, PmxLib.RamAddrList.ErrorMotor)

        return status, motStaData  



    def getRamAccessStatus(self, id: int):
        """
        指定した ID のRAMアクセスステータスを取得します。

        @details MemREADを使用して指定したIDからRAMアクセスステータスを取得します。
        @details 詳しくは取扱説明書をご覧ください。<A HREF= https://kondo-robot.com/faq/pmx-servo-series-online-manual>PMXのオンラインマニュアル</A>
        
        @param id PMXサーボモータのID番号
        
        @retval [status, ramStaData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval ramStaData アクセスエラーになったRAMの先頭アドレス

        @attention この関数を呼ぶとサーボ内のエラー値はリセットされます。
        """

        status, ramStaData = self.MemREADToUint16(id, PmxLib.RamAddrList.ErrorRamAccess)

        return status, ramStaData    



    def getFullStatus(self, id: int):
        """
        指定した ID の全てのステータス情報を取得します。

        @details MemREADを使用して指定したIDから全てのステータスを取得します。
        @details 詳しくは取扱説明書をご覧ください。<A HREF= https://kondo-robot.com/faq/pmx-servo-series-online-manual>PMXのオンラインマニュアル</A>
        
        @param id PMXサーボモータのID番号
        
        @retval [status, servoStatus]
        @retval status 通信エラーとサーボモータのステータスをたしたもの
        @retval servoStatus [staData、sysStaData、motStaData、ramStaData] データが順に含まれます。

        @attention この関数を呼ぶとサーボ内のエラー値はリセットされます。
        """

        status, fulStaData = self.MemREAD(id, PmxLib.RamAddrList.ErrorStatus, 6)

        # 通信失敗の時はデータがないので何も返せない
        if fulStaData is None:
            return status, []
        if len(fulStaData) < 6:  # データが6byte来ないときはうまく通信していない
            return status, []

        servoStatus = []
        
        # statusError
        staData = int.from_bytes(fulStaData[0:1], byteorder='little', signed=False)  # 1byte のみ
        servoStatus.append(staData)

        # SystemError
        sysStaData = int.from_bytes(fulStaData[1:2], byteorder='little', signed=False)  # 1byte のみ
        servoStatus.append(sysStaData)

        # ServoError
        motStaData = int.from_bytes(fulStaData[2:3], byteorder='little', signed=False)  # 1byte のみ
        servoStatus.append(motStaData)

        # RAM Access Error
        ramStaData = int.from_bytes(fulStaData[4:6], byteorder='little', signed=False)  # 2byte
        servoStatus.append(ramStaData)

        return status, servoStatus



    def getTorqueSwitch(self, id: int):
        """
        指定した ID のトルク ON 値を取得します。

        @details MemREADを使用して指定したIDからトルク ON 値を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, trqSwitchData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval trqSwitchData FreeやTorqueOnの値(TorqueSwitchTypeを参照)
        """

        status, trqSwitchData = self.MemREADToByte(id, PmxLib.RamAddrList.TorqueSwitch)

        return status, trqSwitchData



    def getControlMode(self, id: int):
        """
        指定した ID の制御モードを取得します。

        @details MemREADを使用して指定したIDから制御モードを取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 制御モード値。ControlModeを参照してください。
        """

        status, data = self.MemREADToByte(id, PmxLib.RamAddrList.ControlMode)

        return status, data



    def getMotorReceive(self, id: int):
        """
        指定した ID の応答データ指定値を取得します。

        @details MemREADを使用して指定したIDから応答データ指定値を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 応答データ指定値(ReceiveDataOptionを参照)
        """

        status, data = self.MemREADToByte(id, PmxLib.RamAddrList.MotorReceiveData)

        return status, data



    def getTrajectory(self, id: int):
        """
        指定した ID の補間制御軌道生成タイプ指定を取得します。
        
        @details MemREADを使用して指定したIDから補間制御軌道生成タイプ指定を取得します。

        @param id PMXサーボモータのID番号
        
        @retval [status, traData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval traData 補間制御軌道生成タイプ指定。TrajectoryTypeを参照してください。
        """

        status, traData = self.MemREADToByte(id, PmxLib.RamAddrList.Trajectory)

        return status, traData



    def getCenterOffset(self, id: int):
        """
        指定した ID の中央値オフセット値を取得します。

        @details MemREADを使用して指定したIDから中央値オフセット値を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 中央値オフセット値
        """

        status, data = self.MemREADToInt16(id, PmxLib.RamAddrList.CenterOffset)

        return status, data



    def getReverse(self, id: int):
        """
        指定した ID のリバース状態を取得します。

        @details MemREADを使用して指定したIDからクローン/リバースを取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, reverse]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval reverse (bool) リバースON:True,リバースON:False
        
        @note クローン/リバースについては取扱説明書をご覧ください。<A HREF= https://www.kondokagaku.jp/wp_mn/archives/1144#i-39>クローン・リバース設定</A>
        @warning getCloneReverseをしたとしても値は書けますが、Save、再起動を行わないと反映されません
        """

        status, data = self.getCloneReverse(id)

        if status != 0:
            return status, None

        if (data and self.CloneReverseType.Reverse) == self.CloneReverseType.Reverse:
            return status, True
        else:
            return status, False



    def getCloneReverse(self, id: int):
        """
        指定した ID のクローン/リバース値を取得します。

        @details MemREADを使用して指定したIDからクローン/リバースを取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data クローン/リバース値
        
        @note クローン/リバースについては取扱説明書をご覧ください。<A HREF= https://www.kondokagaku.jp/wp_mn/archives/1144#i-39>クローン・リバース設定</A>
        @warning getCloneReverseをしたとしても値は書けますが、Save、再起動を行わないと反映されません
        """

        status, data = self.MemREADToByte(id, PmxLib.RamAddrList.CloneReverse)

        return status, data



    def getMinVoltageLimit(self, id: int):
        """
        指定したIDの下限電圧リミット値を取得します。
        
        @details MemREADを使用して指定したIDから下限電圧リミット値を取得します。

        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 下限電圧設定値
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.MinVoltageLimit)

        return status, data



    def getMinVoltageLimitPower(self, id: int):
        """
        指定したIDの下限リミット電圧を下回った場合の出力割合を取得します。

        @details MemREADを使用して指定したIDから下限リミット電圧を下回った場合の出力割合を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 下限リミット電圧を下回った時の出力割合(%)
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.MinVoltageLimitPower)

        return status, data



    def getMaxVoltageLimit(self, id: int):
        """
        指定したIDの上限リミット電圧を取得します。

        @details MemREADを使用して指定したIDから上限リミット電圧を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 入力電圧最大値
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.MaxVoltageLimit)

        return status, data



    def getMaxVoltageLimitPower(self, id: int):
        """
        指定したIDの上限リミット電圧を超えた場合の出力割合を取得します。

        @details MemREADを使用して指定したIDから上限リミット電圧を超えた場合の出力割合を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 上限リミット電圧を上回った時の出力割合(%)
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.MaxVoltageLimitPower)

        return status, data


    
    def getCurrentLimit(self, id: int):
        """
        指定したIDの電流リミット値を取得します。

        @details MemREADを使用して指定したIDから電流リミット値を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 電流リミット値
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.CurrentLimit)

        return status, data



    def getCurrentLimitPower(self, id: int):
        """
        指定したIDの電流リミットを超えた時の出力割合を取得します。

        @details MemREADを使用して指定したIDから電流リミットを超えた時の出力割合を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 電流リミット時の出力割合(%)
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.CurrentLimitPower)

        return status, data



    def getMotorTempLimit(self, id: int):
        """
        指定したIDのモータ温度リミット値を取得します。

        @details MemREADを使用して指定したIDからモータ温度リミット値を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data モータ温度リミット値
        """

        status, data = self.MemREADToInt16(id, PmxLib.RamAddrList.MotorTempLimit)

        return status, data


    
    def getMotorTempLimitPower(self, id: int):
        """
        指定したIDのモータ温度リミットを超えた時の出力割合を取得します。

        @details MemREADを使用して指定したIDからモータ温度リミットを超えた時の出力割合を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data モータ温度リミット時の出力割合(%)
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.MotorTempLimitPower)

        return status, data


    
    def getCpuTempLimit(self, id: int):
        """
        指定したIDのCPU温度リミット値を取得します。

        @details MemREADを使用して指定したIDからCPU温度リミット値を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data CPU温度リミット値
        """

        status, data = self.MemREADToInt16(id, PmxLib.RamAddrList.CpuTempLimit)

        return status, data



    def getCpuTempLimitPower(self, id: int):
        """
        指定したIDのCPU温度リミットを超えた時の出力割合を取得します。

        @details MemREADを使用して指定したIDからCPU温度リミットを超えた時の出力割合を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data CPU温度リミット時の出力割合(%)
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.CpuTempLimitPower)

        return status, data



    def getCwPositionLimit(self, id: int):
        """
        指定したIDのCW方向の動作角度リミット値を取得します。

        @details MemREADを使用して指定したIDからCW方向の動作角度リミット値を取得します。

        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data CW方向の動作角度リミット値
        """

        status, data = self.MemREADToInt16(id, PmxLib.RamAddrList.CwPositionLimit)

        return status, data



    def getCwPositionLimitPower(self, id: int):
        """
        指定したIDのCW側の角度を超えた時の出力の割合を取得します。

        @details MemREADを使用して指定したIDからCW側の角度を超えた時の出力の割合を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data CW方向の閾値外時の出力割合(%)
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.CwPositionLimitPower)

        return status, data



    def getCcwPositionLimit(self, id: int):
        """
        指定したIDのCCW方向の動作角度リミット値を取得します。

        @details MemREADを使用して指定したIDからCCW方向の動作角度リミット値を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data CW方向の動作角度リミット値
        """

        status, data = self.MemREADToInt16(id, PmxLib.RamAddrList.CcwPositionLimit)

        return status, data



    def getCcwPositionLimitPower(self, id: int):
        """
        指定したIDのCCW側の角度を超えた時の出力の割合を取得します。

        @details MemREADを使用して指定したIDからCCW側の角度を超えた時の出力の割合を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data CCW方向の閾値外時の出力割合(%)
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.CcwPositionLimitPower)

        return status, data



    def getMaxGoalSpeed(self, id: int):
        """
        指定したIDの指令可能な最大速度値を取得します。

        @details MemREADを使用して指定したIDから指令可能な最大速度値を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 最大速度指令値
        
        @attention 最大速度指令値を超えて速度指示した場合はエラーが返り、実行はされません。
        """

        status, data = self.MemREADToInt16(id, PmxLib.RamAddrList.MaxGoalSpeed)

        return status, data



    def getMaxGoalCurrent(self, id: int):
        """
        指定したIDの指令可能な最大電流値を取得します。

        @details MemREADを使用して指定したIDから指令可能な最大電流値を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 最大電流指令値
        
        @attention 最大電流指令値を超えて電流指示した場合はエラーが返り、実行はされません。
        """

        status, data = self.MemREADToInt16(id, PmxLib.RamAddrList.MaxGoalCurrent)

        return status, data



    def getMaxGoalTorque(self, id: int):
        """
        指定したIDの指令可能な最大推定トルク値を取得します。
        
        @details MemREADを使用して指定したIDから指令可能な最大推定トルク値を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data 最大動推定トルク指令値
        
        @attention 最大推定トルク指令値を超えてトルク指示した場合はエラーが返り、実行はされません。
        """

        status, data = self.MemREADToInt16(id, PmxLib.RamAddrList.MaxGoalTorque)

        return status, data



    def getTotalPowerRate(self, id: int):
        """
        指定したIDのモータに出力する割合を取得します。

        @details MemREADを使用して指定したIDからモータに出力する割合を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data モータ出力割合(%)
        
        @note 出力のリミットではなく全体的にモータ出力を可変させます。
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.TotalPowerRate)

        return status, data


    
    def getLockDetectTime(self, id: int):
        """
        指定したIDのロック検知を認識するまでの時間を取得します。

        @details MemREADを使用して指定したIDからロック検知を認識するまでの時間を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data ロック検知を認識するまでの時間
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.LockDetectTime)

        return status, data


    
    def getLockThresholdPower(self, id: int):
        """
        指定したIDのロック検知を開始する出力指示の閾値を取得します。

        @details MemREADを使用して指定したIDからロック検知を開始する出力指示の閾値を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data ロック検知開始出力(%)
        
        @note 位置制御モードのみで位置制御を含まないモードの場合は使用しません。
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.LockThresholdPower)

        return status, data


    
    def getLockDetectOutputPower(self, id: int):
        """
        指定したIDがロック検知が起きた時の出力の割合を取得します。

        @details MemREADを使用して指定したIDからロック検知が起きた時の出力の割合を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data ロック検知時の出力割合(%)
        """

        status, data = self.MemREADToUint16(id, PmxLib.RamAddrList.LockDetectOutputPower)

        return status, data



    def getCenterOffsetRange(self, id: int):
        """
        指定したIDの中央値オフセット(CenterOffset)の設定可能な最小/最大の範囲を取得します。

        @details MemREADを使用して指定したIDから中央値オフセット(CenterOffset)の設定可能な最小/最大の範囲を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, offsetData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval offsetData 中央値オフセットの範囲([CenterOffsetの最小値,CenterOffsetの最大値])
        """
        
        offsetData = []
        
        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.CenterOffsetMinRange, 4)

        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray[0:2], byteorder='little', signed=True)  # int16 リトルエンディアン
            
            offsetData.append(reData)
            reData = int.from_bytes(byteArray[2:4], byteorder='little', signed=True)  # int16 リトルエンディアン
            offsetData.append(reData)

        return status, offsetData



    def getMinVoltageLimitRange(self, id: int):
        """
        指定したIDの入力最小電圧リミット値(MinVoltageLimit)の設定可能な最小/最大の範囲を取得します。

        @details MemREADを使用して指定したIDから入力最小電圧リミット値(MinVoltageLimit)の設定可能な最小/最大の範囲を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, rangeData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rangeData 入力最小電圧リミット値の範囲([MinVoltageLimitの最小値,MinVoltageLimitの最大値])
        """
        
        rangeData = []
        
        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.MinVoltageMinRange, 4)
        
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray[0:2], byteorder='little', signed=False)  # Uint16 リトルエンディアン
            rangeData.append(reData)
            reData = int.from_bytes(byteArray[2:4], byteorder='little', signed=False)  # Uint16 リトルエンディアン
            rangeData.append(reData)

        return status, rangeData



    def getMaxVoltageLimitRange(self, id: int):
        """
        指定した ID の最大電圧(MaxVoltage)の設定可能な範囲を取得します。

        @details MemREADを使用して指定したIDから最大電圧(MaxVoltage)の範囲を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, rangeData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rangeData 最大電圧の範囲([MaxVoltageの最小値,MaxVoltageの最大値])
        """
        
        rangeData = []
        
        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.MaxVoltageMinRange, 4)
        
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray[0:2], byteorder='little', signed=False)  # Uint16 リトルエンディアン
            rangeData.append(reData)
            reData = int.from_bytes(byteArray[2:4], byteorder='little', signed=False)  # Uint16 リトルエンディアン
            rangeData.append(reData)

        return status, rangeData



    def getCurrentLimitRange(self, id: int):
        """
        指定した ID の電流(Current)の設定可能な範囲を取得します。

        @details MemREADを使用して指定したIDから電流(Current)の範囲を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, rangeData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rangeData 電流の範囲([Currentの最小値,Currentの最大値])
        """
        
        rangeData = []
        
        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.CurrentMinRange, 4)
        
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray[0:2], byteorder='little', signed=False)  # Uint16 リトルエンディアン
            rangeData.append(reData)
            reData = int.from_bytes(byteArray[2:4], byteorder='little', signed=False)  # Uint16 リトルエンディアン
            rangeData.append(reData)

        return status, rangeData



    def getMotorTempLimitRange(self, id: int):
        """
        指定した ID のモータ温度(MotorTemp)の範囲を取得します。

        @details MemREADを使用して指定したIDからモータ温度(MotorTemp)の設定可能な範囲を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, rangeData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rangeData モータ温度の範囲([MotorTempの最小値,MotorTempの最大値])
        """
        
        rangeData = []
        
        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.MotorTempMinRange, 4)
        
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray[0:2], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)
            reData = int.from_bytes(byteArray[2:4], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)

        return status, rangeData



    def getCpuTempLimitRange(self, id: int):
        """
        指定した ID のCPU温度の範囲(CpuTemp)を取得します。

        @details MemREADを使用して指定したIDからCPU温度(CpuTemp)の設定可能な範囲を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, rangeData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rangeData CPU温度の範囲([CpuTempの最小値,CpuTempの最大値])
        """
        
        rangeData = []
        
        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.CpuTempMinRange, 4)
        
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray[0:2], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)
            reData = int.from_bytes(byteArray[2:4], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)

        return status, rangeData



    def getMaxGoalSpeedRange(self, id: int):
        """
        指定した ID の最大速度指令値(MaxGoalSpeed)の設定可能な範囲を取得します。

        @details MemREADを使用して指定したIDから最大速度指令値(MaxGoalSpeed)の範囲を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, rangeData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rangeData 最大速度指令値の範囲([MaxGoalSpeedの最小値,MaxGoalSpeedの最大値])
        """
        
        rangeData = []
        
        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.MaxGoalSpeedMinRange, 4)
        
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray[0:2], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)
            reData = int.from_bytes(byteArray[2:4], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)

        return status, rangeData



    def getMaxGoalCurrentRange(self, id: int):
        """
        指定した ID の最大電流指令値(MaxGoalCurrent)の設定可能な範囲を取得します。

        @details MemREADを使用して指定したIDから最大電流指令値(MaxGoalCurrent)の範囲を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, rangeData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rangeData 最大電流指令値の範囲([MaxGoalCurrentの最小値,MaxGoalCurrentの最大値])
        """
        
        rangeData = []
        
        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.MaxGoalCurrentMinRange, 4)
        
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray[0:2], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)
            reData = int.from_bytes(byteArray[2:4], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)

        return status, rangeData



    def getMaxGoalTorqueRange(self, id: int):
        """
        指定した ID の最大トルク指令値(MaxGoalTorque)の設定可能な範囲を取得します。

        @details MemREADを使用して指定したIDから最大トルク指令値(MaxGoalTorque)の範囲を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, rangeData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rangeData 最大トルク指令値の範囲([MaxGoalTorqueの最小値,MaxGoalTorqueの最大値])
        """
        
        rangeData = []
        
        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.MaxGoalTorqueMinRange, 4)
        
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray[0:2], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)
            reData = int.from_bytes(byteArray[2:4], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)

        return status, rangeData



    def getCwPositionLimitRange(self, id: int):
        """
        指定したIDの指定したCW方向の動作角度リミット値(CwPosition)の設定可能な範囲を取得します。

        @details MemREADを使用して指定したIDから指定したCW方向の動作角度リミット値(CwPosition)の範囲を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, rangeData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rangeData 最大位置の範囲([CwPositionの最小値,CwPositionの最大値])
        """
        
        rangeData = []
        
        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.CwPositionMinRange, 4)
        
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray[0:2], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)
            reData = int.from_bytes(byteArray[2:4], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)

        return status, rangeData



    def getCcwPositionLimitRange(self, id: int):
        """
        指定したIDの指定したCCW方向の動作角度リミット値(CcwPosition)の設定可能な範囲を取得します。

        @details MemREADを使用して指定したIDから指定したCCW方向の動作角度リミット値(CcwPosition)の範囲を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, rangeData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval rangeData 最小位置の範囲([CcwPositionの最小値,CcwPositionの最大値])
        """
        
        rangeData = []
        
        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.CcwPositionMinRange, 4)
        
        if len(byteArray) == 4:
            reData = int.from_bytes(byteArray[0:2], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)
            reData = int.from_bytes(byteArray[2:4], byteorder='little', signed=True)  # int16 リトルエンディアン
            rangeData.append(reData)

        return status, rangeData



    def printAllRange(self, id: int):
        """
        指定した ID の設定データの範囲を表示する関数
        
        @param id PMXサーボモータのID番号
        @return status 最後に行った送受信結果(送信結果+PMXのstatus(Uint16))
        """
    
        status, offsetRange = self.getCenterOffsetRange(id)
        print("中央値オフセット設定範囲(600)[x100deg]>>", hex(status), offsetRange)
    
        status, minVolLimitRange = self.getMinVoltageLimitRange(id)
        print("下限電圧リミット設定範囲(604)[mV]>>", hex(status), minVolLimitRange)
        
        status, maxVolLimitRange = self.getMaxVoltageLimitRange(id)
        print("上限電圧リミット設定範囲(608)[mV]>>", hex(status), maxVolLimitRange)
        
        status, maxCurLimitRange = self.getCurrentLimitRange(id)
        print("電流リミット設定範囲(616)[mA]>>", hex(status), maxCurLimitRange)
        
        status, motorTempLimitRange = self.getMotorTempLimitRange(id)
        print("モータ温度リミット設定範囲(620)[x10℃]>>", hex(status), motorTempLimitRange)
        
        status, cpuTempLimitRange = self.getCpuTempLimitRange(id)
        print("CPU温度リミット設定範囲(624)[x10℃]>>", hex(status), cpuTempLimitRange)
        
        status, cwPosRange = self.getCwPositionLimitRange(id)
        print("CW方向設定範囲(628)[x100deg]>>", hex(status), cwPosRange)
        
        status, ccwPosRange = self.getCcwPositionLimitRange(id)
        print("CCW方向設定範囲(632)[x100deg]>>", hex(status), ccwPosRange)
    
        status, goalSpdRange = self.getMaxGoalSpeedRange(id)
        print("最大速度指令値設定範囲(636)[x10deg/s]>>", hex(status), goalSpdRange)
        
        status, goalCurRange = self.getMaxGoalCurrentRange(id)
        print("最大電流指令値設定範囲(640)[mA]>>", hex(status), goalCurRange)
        
        status, goalTrqRange = self.getMaxGoalTorqueRange(id)
        print("最大トルク指令値設定範囲(644)[mN.m]>>", hex(status), goalTrqRange)

        # 最後に読み取ったステータスを返す
        return status



    def getDafaultPidGain(self,id: int,pAddr :int):
        """
        指定した ID の PIDゲインを取得します。
        
        @param id PMXサーボモータのID番号
        @param pAddr PIDゲインのkpが保存されている先頭のアドレス
        
        @retval [status, pidArray]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval pidArray 読み取ったPIDゲインのデータ配列[kp, ki, kd, ks]。通信失敗時には空のリストを返します。
        """
    
        status, byteArray = self.MemREAD(id, pAddr, 16)

        if (status == PmxLib.ComError.TimeOut):
            return PmxLib.ComError.TimeOut, []
        
        elif (len(byteArray) == 16 ):
            kp = int.from_bytes(byteArray[0:4], byteorder='little', signed=False)  # int16 リトルエンディアン
            ki = int.from_bytes(byteArray[4:8], byteorder='little', signed=False)  # int16 リトルエンディアン
            kd = int.from_bytes(byteArray[8:12], byteorder='little', signed=False)  # int16 リトルエンディアン
            ks = int.from_bytes(byteArray[12:16], byteorder='little', signed=False)  # int16 リトルエンディアン
            pidArray = [kp, ki, kd, ks]
        else:
            return status + PmxLib.ComError.ReceiveError,[]

        return status,pidArray



    def getPresetAllPidGains(self, id: int,presetNum :int):
        """
        指定した ID 、プリセット番号の PID ゲインを位置からトルクまで全て取得します。
        
        @param id PMXサーボモータのID番号
        @param presetNum プリセット番号
        
        @retval [status, gainsList]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval gainsList 読み取ったPIDゲインのデータ配列(16データ)。通信失敗時には空のリストを返します。
        
        @version 1.0.2.0
        
        @remark PMXのV1.1.0.xから有効
        
        """

        startAddr = 0
        
        if presetNum == 1:
            startAddr = PmxLib.RamAddrList.PositionKp
        elif presetNum == 2:
            startAddr = PmxLib.RamAddrList.PositionKp2
        elif presetNum == 3:
            startAddr = PmxLib.RamAddrList.PositionKp3
        else:
            return PmxLib.ComError.FormatError,[]

        status, byteArray = self.MemREAD(id, startAddr, 64)
        
        gainsList = []

        if (status == PmxLib.ComError.TimeOut):
            return PmxLib.ComError.TimeOut, []
        
        elif (len(byteArray) == 64):
            
            for gainCount in range(0,4):

                pidGains = []
                offsetGain = gainCount * 16
                
                for pidCount in range(0,4):
                    pidOffset =  offsetGain + pidCount * 4
                    gain = int.from_bytes(byteArray[pidOffset: pidOffset+4], byteorder='little', signed=False)  # Uint32 リトルエンディアン
                    pidGains.append(gain)

                gainsList.append(pidGains)

        else:
            return status + PmxLib.ComError.ReceiveError, []
        
        return status, gainsList



    def getPresetPidGains(self, id: int, presetNum :int, gainType :int):
        """
        指定した ID 、プリセット番号、ゲインタイプのPID ゲインを取得します。
        
        @param id PMXサーボモータのID番号
        @param presetNum プリセット番号
        @param gainType ゲインタイプ(0:位置制御ゲイン,1:速度制御ゲイン,2:電流制御ゲイン,3:トルク制御ゲイン)
        
        @retval [status, pidArray]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval pidArray 読み取ったPIDゲインのデータ配列[kp, ki, kd, ks]。通信失敗時には空のリストを返します。
        
        @version 1.0.2.0
        
        @remark PMXのV1.1.0.xから有効
        
        """
        
        # ゲインタイプの確認
        if gainType < 0 or 3 < gainType:
            return PmxLib.ComError.FormatError,[]

        startAddr = 0
        
        if presetNum == 1:
            if gainType == 0:
                startAddr = PmxLib.RamAddrList.PositionKp
            elif gainType == 1:
                startAddr = PmxLib.RamAddrList.SpeedKp
            elif gainType == 2:
                startAddr = PmxLib.RamAddrList.CurrentKp
            elif gainType == 3:
                startAddr = PmxLib.RamAddrList.TorqueKp
        elif presetNum == 2:
            if gainType == 0:
                startAddr = PmxLib.RamAddrList.PositionKp2
            elif gainType == 1:
                startAddr = PmxLib.RamAddrList.SpeedKp2
            elif gainType == 2:
                startAddr = PmxLib.RamAddrList.CurrentKp2
            elif gainType == 3:
                startAddr = PmxLib.RamAddrList.TorqueKp2
        elif presetNum == 3:
            if gainType == 0:
                startAddr = PmxLib.RamAddrList.PositionKp3
            elif gainType == 1:
                startAddr = PmxLib.RamAddrList.SpeedKp3
            elif gainType == 2:
                startAddr = PmxLib.RamAddrList.CurrentKp3
            elif gainType == 3:
                startAddr = PmxLib.RamAddrList.TorqueKp3
        else:
            return PmxLib.ComError.FormatError,[]
        
        return self.getDafaultPidGain(id, startAddr)



    def getAllPresetNum(self, id: int):
        """
        指定した ID のすべてのプリセット番号を取得します。

        @param id PMXサーボモータのID番号

        @retval [status, presetArray]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval presetArray 読み取ったプリセット番号のデータ配列[位置ゲインプリセット, 速度ゲインプリセット, 電流ゲインプリセット, トルクゲインプリセット]。通信失敗時には空のリストを返します。
        
        @version 1.0.2.0
        
        @remark PMXのV1.1.0.xから有効
        
        """

        status, byteArray = self.MemREAD(id, PmxLib.RamAddrList.presetPosAddr, 4)

        presetArray = []

        if (status == PmxLib.ComError.TimeOut):
            return PmxLib.ComError.TimeOut, []
        
        elif (len(byteArray) == 4 ):
            pPos = int.from_bytes(byteArray[0:1], byteorder='little', signed=False)  # byte リトルエンディアン
            pSpd = int.from_bytes(byteArray[1:2], byteorder='little', signed=False)  # byte リトルエンディアン
            pCur = int.from_bytes(byteArray[2:3], byteorder='little', signed=False)  # byte リトルエンディアン
            pTrq = int.from_bytes(byteArray[3:4], byteorder='little', signed=False)  # byte リトルエンディアン
            presetArray = [pPos,pSpd,pCur,pTrq]

        else:
            return status+ PmxLib.ComError.ReceiveError,[]

        return status,presetArray



    def getLedMode(self, id: int):
        """
        指定したID のLED点灯モードを取得します。

        @param id PMXサーボモータのID番号
        
        @retval [status, data]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval data LED点灯モード
        
        @version 1.0.2.0
        
        @remark PMXのV1.1.0.xから有効
        
        """
        status, data = self.MemREADToByte(id, PmxLib.RamAddrList.LedMode)

        return status, data



    #################################################################################################################    
    
    #################################################################################################################
    # MemWriteを利用したデータのset関数一覧
    #################################################################################################################
    
    
    def setTorqueSwitch(self, id: int, data: int, writeOpt=0):
        """
        MemWRITEを試用したトルクスイッチ。 IDで指定したサーボにTorqueON、Free、Brake、Holdを指定することができます
        
        @param id PMXサーボモータのID番号
        @param data byte(符号なし1byte)の書きこみデータ
        @param writeOpt MemWRITEで使用するオプション 0:通常書き込み、1:TorqueOn中の強制書き込み

        @return status MemWRITEの送受信結果(送信結果+PMXのstatus(Uint16))
        
        @version 1.0.1.0～
        
        """

        return self.MemWRITEToByte(id, PmxLib.RamAddrList.TorqueSwitch, data, writeOpt=writeOpt)



    def setMotorReceive(self, id: int, receiveOpt: int, writeOpt=0):
        """
        指定した ID のmotorREADやMotorWRITEから返信される応答モードを設定します。
        
        @param id PMXサーボモータのID番号
        @param receiveOpt 応答モード ReceiveDataOptionを参照 
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)

        @return status 送受信結果(送信結果+PMXのstatus(Uint16))
        """
        
        return self.MemWRITEToByte(id, PmxLib.RamAddrList.MotorReceiveData, receiveOpt, writeOpt=writeOpt)  # id,addr,byteArray



    def setControlMode(self, id: int, controlMode: int, writeOpt=0):
        """
        指定した ID の位置制御などの制御モードを設定します。
        
        @param id PMXサーボモータのID番号
        @param controlMode モータに設定する制御モード(ControlModeを参照)
        @param writeOpt writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return status 送受信結果(送信結果+PMXのstatus(Uint16))
        
        @version 1.0.1.0
        - writeOptが反映されない不具合を修正しました
        """

        return self.MemWRITEToByte(id, PmxLib.RamAddrList.ControlMode, controlMode, writeOpt=writeOpt)  # id,addr,byteArray
    


    def setDafaultPidGain(self, id: int, pidAddr: int, kp: int, ki: int, kd: int, st=None, writeOpt=0):
        """
        指定した ID のPIDゲインを変更します。
        
        @param id PMXサーボモータのID番号
        @param pidAddr PIDゲインのkpが保存されている先頭のアドレス
        @param kp Pゲイン
        @param ki Iゲイン
        @param kd Dゲイン
        @param st ストレッチゲイン
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        """

        sendData = []
        
        sendData.extend(kp.to_bytes(4, byteorder="little", signed=False))
        sendData.extend(ki.to_bytes(4, byteorder="little", signed=False))
        sendData.extend(kd.to_bytes(4, byteorder="little", signed=False))

        if st is not None:
            sendData.extend(st.to_bytes(4, byteorder="little", signed=False))
        
        return self.MemWRITE(id, pidAddr, sendData, writeOpt=writeOpt)



    def setPositionPidGain(self, id: int, kp: int, ki: int, kd: int, st: int, writeOpt=0):
        """
        指定した ID のプリセット1の位置制御のPIDゲインを変更します。
        
        @param id PMXサーボモータのID番号
        @param kp 位置制御のPゲイン
        @param ki 位置制御のIゲイン
        @param kd 位置制御のDゲイン
        @param st 位置のストレッチゲイン
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        """

        pidPAddr = PmxLib.RamAddrList.PositionKp
        return self.setDafaultPidGain(id, pidPAddr, kp, ki, kd, st=st, writeOpt=writeOpt)



    def setSpeedPidGain(self, id: int, kp: int, ki: int, kd: int, writeOpt=0):
        """
        指定した ID のプリセット1の速度制御のPIDゲインを変更します。
        
        @param id PMXサーボモータのID番号
        @param kp 速度制御のPゲイン
        @param ki 速度制御のIゲイン
        @param kd 速度制御のDゲイン
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        """

        pidPAddr = PmxLib.RamAddrList.SpeedKp
        return self.setDafaultPidGain(id, pidPAddr, kp, ki, kd, writeOpt=writeOpt)



    def setCurrentPidGain(self, id: int, kp: int, ki: int, kd: int, writeOpt=0):
        """
        指定した ID のプリセット1の電流制御のPIDゲインを変更します。
        
        @param id PMXサーボモータのID番号
        @param kp 電流制御のPゲイン
        @param ki 電流制御のIゲイン
        @param kd 電流制御のDゲイン
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        """

        pidPAddr = PmxLib.RamAddrList.CurrentKp
        return self.setDafaultPidGain(id, pidPAddr, kp, ki, kd, writeOpt=writeOpt)



    def setTorquePidGain(self, id: int, kp: int, ki: int, kd: int, writeOpt=0):
        """
        指定した ID のプリセット1のトルク制御のPIDゲインを変更します。
        
        @param id PMXサーボモータのID番号
        @param kp トルク制御のPゲイン
        @param ki トルク制御のIゲイン
        @param kd トルク制御のDゲイン
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        """

        pidPAddr = PmxLib.RamAddrList.TorqueKp
        return self.setDafaultPidGain(id, pidPAddr, kp, ki, kd, writeOpt=writeOpt)



    def setCenterOffset(self, id: int, CenterOffset: int, writeOpt=0):
        """
        指定した ID の中央値オフセット値を設定します。
        
        @param id PMXサーボモータのID番号
        @param CenterOffset 中央値オフセット値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)

        @return 送受信結果(送信結果+PMXのstatus(Uint16))

        @attention 範囲外のデータを受け取った場合statusにエラーが返ってきます。
        @attention エラー時は実行されません。
        """

        return self.MemWRITEToInt16(id, PmxLib.RamAddrList.CenterOffset, CenterOffset, writeOpt=writeOpt)



    def setCloneReverse(self, id: int, crVal: int):
        """
        指定した ID のクローン/リバースを設定します。
        
        @param id PMXサーボモータのID番号
        @param crVal クローン/リバースを設定する値。クローンなら1、リバースなら2を入力します。
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        
        @note クローン/リバースについては取扱説明書をご覧ください。<A HREF= https://www.kondokagaku.jp/wp_mn/archives/1144#i-39>クローン・リバース設定</A>
        @attention 強制書き込みをすると危ないのでoptは使用しません。
        @attention reverse/cloneを設定した場合はSAVEをして再起動をかけてください。
        """
        
        return self.MemWRITEToByte(id, PmxLib.RamAddrList.CloneReverse, crVal)



    def setMinVoltageLimit(self, id: int, minVol: int, limPower: int, writeOpt=0):
        """
        指定した ID の下限電圧を下回った時の挙動を設定します。
        
        @param id PMXサーボモータのID番号
        @param minVol 入力電圧最小値
        @param limPower 入力電圧最小時の出力％値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)

        @return 送受信結果(送信結果+PMXのstatus(Uint16))

        @attention 範囲外のデータを受け取った場合statusにエラーが返ってきます。
        @attention エラー時は実行されません。
        """
        
        sendData = []
        
        sendData.extend(minVol.to_bytes(2, byteorder="little", signed=False))
        sendData.extend(limPower.to_bytes(2, byteorder="little", signed=False))

        return self.MemWRITE(id, PmxLib.RamAddrList.MinVoltageLimit, sendData, writeOpt=writeOpt)



    def setMaxVoltageLimit(self, id: int, maxVol: int, limPower: int, writeOpt=0):
        """
        指定した ID の上限電圧を上回った時の挙動を設定します。
        
        @param id PMXサーボモータのID番号
        @param maxVol 入力電圧最大値
        @param limPower 入力電圧最大時の出力％値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))

        @attention 範囲外のデータを受け取った場合statusにエラーが返ってきます。
        @attention エラー時は実行されません。
        """

        sendData = []
        
        sendData.extend(maxVol.to_bytes(2, byteorder="little", signed=False))
        sendData.extend(limPower.to_bytes(2, byteorder="little", signed=False))

        return self.MemWRITE(id, PmxLib.RamAddrList.MaxVoltageLimit, sendData, writeOpt=writeOpt)



    def setCurrentLimit(self, id: int, maxCur: int, limPower: int, writeOpt=0):
        """
        指定した ID のモータ消費電流最大値を上回った時の挙動を設定します。
        
        @param id PMXサーボモータのID番号
        @param maxCur モータ消費電流最大値
        @param limPower モータ消費電流最大時の出力％値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))

        @attention 範囲外のデータを受け取った場合statusにエラーが返ってきます。
        @attention エラー時は実行されません。
        """

        sendData = []
        
        sendData.extend(maxCur.to_bytes(2, byteorder="little", signed=False))
        sendData.extend(limPower.to_bytes(2, byteorder="little", signed=False))

        return self.MemWRITE(id, PmxLib.RamAddrList.CurrentLimit, sendData, writeOpt=writeOpt)



    def setMotorTempLimit(self, id: int, maxTemp: int, limPower: int, writeOpt=0):
        """
        指定した ID のモータ温度最大値を上回った時の挙動を設定します。

        @param id PMXサーボモータのID番号
        @param maxTemp モータ温度最大値
        @param limPower モータ温度最大時の出力％値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)

        @return 送受信結果(送信結果+PMXのstatus(Uint16))

        @attention 範囲外のデータを受け取った場合statusにエラーが返ってきます。
        @attention エラー時は実行されません。
        """
            
        sendData = []
        
        sendData.extend(maxTemp.to_bytes(2, byteorder="little", signed=True))
        sendData.extend(limPower.to_bytes(2, byteorder="little", signed=False))

        return self.MemWRITE(id, PmxLib.RamAddrList.MotorTempLimit, sendData)



    def setCpuTempLimit(self, id: int, maxTemp: int, limPower: int, writeOpt=0):
        """
        指定した ID のCPU温度最大値を上回った時の挙動を設定します。
        
        @param id PMXサーボモータのID番号
        @param maxTemp CPU温度最大値
        @param limPower CPU温度最大時の出力％値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)

        @return 送受信結果(送信結果+PMXのstatus(Uint16))

        @attention 範囲外のデータを受け取った場合statusにエラーが返ってきます。
        @attention エラー時は実行されません。
        """

        sendData = []
        
        sendData.extend(maxTemp.to_bytes(2, byteorder="little", signed=True))
        sendData.extend(limPower.to_bytes(2, byteorder="little", signed=False))

        return self.MemWRITE(id, PmxLib.RamAddrList.CpuTempLimit, sendData)



    def setPositionLimit(self, id: int, cwPos: int, ccwPos: int, limitPower: int, writeOpt=0):
        """
        指定した ID の最大角を上回った時の挙動を設定します。
        
        @param id PMXサーボモータのID番号
        @param cwPos CW方向最大角
        @param ccwPos CCW方向最大角
        @param limitPower 最大角閾値外時の出力％値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)

        @return 送受信結果(送信結果+PMXのstatus(Uint16))

        @attention 範囲外のデータを受け取った場合statusにエラーが返ってきます。
        @attention エラー時は実行されません。
        """
        
        sendData = []
        
        sendData.extend(cwPos.to_bytes(2, byteorder="little", signed=True))
        sendData.extend(limitPower.to_bytes(2, byteorder="little", signed=False))
        sendData.extend(ccwPos.to_bytes(2, byteorder="little", signed=True))
        sendData.extend(limitPower.to_bytes(2, byteorder="little", signed=False))

        return self.MemWRITE(id, PmxLib.RamAddrList.CwPositionLimit, sendData, writeOpt=writeOpt)



    def setMaxGoalSpeed(self, id: int, maxDesSpd: int, writeOpt=0):
        """
        指定した ID の最大速度指令値を設定します。
        
        @param id PMXサーボモータのID番号
        @param maxDesSpd 最大速度指令値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))

        @attention 範囲外のデータを受け取った場合statusにエラーが返ってきます。
        @attention エラー時は実行されません。
        """
        
        return self.MemWRITEToInt16(id, PmxLib.RamAddrList.MaxGoalSpeed, maxDesSpd, writeOpt=writeOpt)



    def setMaxGoalCurrent(self, id: int, maxDesCur: int, writeOpt=0):
        """
        指定した ID の最大電流指令値を設定します。

        @param id PMXサーボモータのID番号
        @param maxDesCur 最大電流指令値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))

        @attention 範囲外のデータを受け取った場合statusにエラーが返ってきます。
        @attention エラー時は実行されません。
        """

        return self.MemWRITEToInt16(id, PmxLib.RamAddrList.MaxGoalCurrent, maxDesCur, writeOpt=writeOpt)



    def setMaxGoalTorque(self, id: int, maxDesTrq: int, writeOpt=0):
        """
        指定した ID の最大動推定トルク指令値を設定します。

        @param id PMXサーボモータのID番号
        @param maxDesTrq 最大動推定トルク指令値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))

        @attention 範囲外のデータを受け取った場合statusにエラーが返ってきます。
        @attention エラー時は実行されません。
        """

        return self.MemWRITEToInt16(id, PmxLib.RamAddrList.MaxGoalTorque, maxDesTrq, writeOpt=writeOpt)



    def setTotalPowerRate(self, id: int, rate: int, writeOpt=0):
        """
        指定した ID のモータ出力制限％値を設定します。

        @param id PMXサーボモータのID番号
        @param rate モータ出力制限％値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        """

        return self.MemWRITEToUint16(id, PmxLib.RamAddrList.TotalPowerRate, rate)
        


    def setLockDetect(self, id: int, time: int, power: int, outputPower: int, writeOpt=0):
        """
        指定した ID のロック時間、ロックと認識される出力割合、およびロック時間の出力％値を設定します。
        
        @param id PMXサーボモータのID番号
        @param time ロック時間
        @param power ロックと認識される出力割合
        @param outputPower ロック時間の出力％値
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        """

        sendData = []
        
        sendData.extend(time.to_bytes(2, byteorder="little", signed=False))
        sendData.extend(power.to_bytes(2, byteorder="little", signed=False))
        sendData.extend(outputPower.to_bytes(2, byteorder="little", signed=False))

        return self.MemWRITE(id, PmxLib.RamAddrList.LockDetectTime, sendData, writeOpt=writeOpt)



    def setTrajectory(self, id: int, data: int, writeOpt=0):
        """
        指定した ID の補間制御軌道生成タイプ指定を設定します。
        
        @param id PMXサーボモータのID番号
        @param data 補間制御軌道生成タイプ指定
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        """
        
        return self.MemWRITEToByte(id, PmxLib.RamAddrList.Trajectory, data, writeOpt=writeOpt)



    def setAllPresetNum(self, id: int,presetNum :int, writeOpt=0):
        """
        指定した ID の全てのプリセット番号を設定します。

        @param id PMXサーボモータのID番号
        @param presetNum プリセット番号
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)

        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        
        @version 1.0.2.0
        
        @remark PMXのV1.1.0.xから有効
        
        """
    
        if presetNum < 0 or 3 < presetNum:
            return PmxLib.ComError.FormatError

        # 位置/速度/電流/トルクすべて同じプリセットを書く
        writeData = [presetNum,presetNum,presetNum,presetNum]

        status = self.MemWRITE(id,PmxLib.RamAddrList.presetPosAddr,writeData,writeOpt)

        return status



    def setPidGains(self, id: int, presetNum:int, gainType :int, kp: int, ki: int, kd: int, st=None,writeOpt=0):
        """
        ID、プリセット番号、ゲインタイプを指定し、PIDゲインを変更します。
        
        @param id PMXサーボモータのID番号
        @param presetNum プリセット番号
        @param gainType ゲインタイプ(0:位置制御ゲイン,1:速度制御ゲイン,2:電流制御ゲイン,3:トルク制御ゲイン)
        @param kp Pゲイン
        @param ki Iゲイン
        @param kd Dゲイン
        @param st ストレッチゲイン
        @param writeOpt TolqueOn状態でも強制的に書き込むか設定します。0:強制書き込みしない、1:強制書き込みする(デフォルトは0)
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        
        @version 1.0.2.0
        
        @remark PMXのV1.1.0.xから有効
        
        """
        # ゲインタイプの確認
        if gainType < 0 or 3 < gainType:
            return PmxLib.ComError.FormatError
        
        startAddr = 0

        if presetNum == 1:
            if gainType == 0:
                startAddr = PmxLib.RamAddrList.PositionKp
            elif gainType == 1:
                startAddr = PmxLib.RamAddrList.SpeedKp
            elif gainType == 2:
                startAddr = PmxLib.RamAddrList.CurrentKp
            elif gainType == 3:
                startAddr = PmxLib.RamAddrList.TorqueKp
        elif presetNum == 2:
            if gainType == 0:
                startAddr = PmxLib.RamAddrList.PositionKp2
            elif gainType == 1:
                startAddr = PmxLib.RamAddrList.SpeedKp2
            elif gainType == 2:
                startAddr = PmxLib.RamAddrList.CurrentKp2
            elif gainType == 3:
                startAddr = PmxLib.RamAddrList.TorqueKp2
        elif presetNum == 3:
            if gainType == 0:
                startAddr = PmxLib.RamAddrList.PositionKp3
            elif gainType == 1:
                startAddr = PmxLib.RamAddrList.SpeedKp3
            elif gainType == 2:
                startAddr = PmxLib.RamAddrList.CurrentKp3
            elif gainType == 3:
                startAddr = PmxLib.RamAddrList.TorqueKp3
        else:
            return PmxLib.ComError.FormatError

        return self.setDafaultPidGain(id, startAddr, kp, ki, kd, st, writeOpt=0)
    


    def setLedMode(self, id: int, ledVal: int):
        """
        指定した ID のLEDモードを設定します。

        @param id PMXサーボモータのID番号
        @param ledVal LEDを設定する値。デフォルトは0、消灯は1、点灯は2を入力します。
        
        @return 送受信結果(送信結果+PMXのstatus(Uint16))
        
        @version 1.0.2.0
        
        @remark PMXのV1.1.0.xから有効
        """
        return self.MemWRITEToByte(id, PmxLib.RamAddrList.LedMode, ledVal)
    



    #################################################################################################################    
    
    #################################################################################################################
    # MotorWRITEを利用したデータのset関数一覧
    #################################################################################################################

    def setMotorTorqueOn(self, id: int, receiveMode: int, controlMode=0x00):
        """
        指定した ID のモータをトルクオンにし、ステータス、モータのON/OFF、受信データを取得します。
        
        @param id PMXサーボモータのID番号
        @param receiveMode PMXで設定した応答モード
        @param controlMode モータに設定する制御モード(ControlModeを参照)

        @retval [status, torqueSwitch, rxDataList]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval torqueSwitch FreeやTorqueOnの値(TorqueSwitchTypeを参照) 
        @retval rxDataList MotorReadで読み取ったデータ[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧](合計8項目)
        """

        return self.MotorWRITE(id, PmxLib.TorqueSwitchType.TorqueOn, controlMode=controlMode, receiveMode=receiveMode)



    def setMotorFree(self, id: int, receiveMode: int, controlMode=0x00):
        """
        指定した ID のモータをフリーにし、ステータス、モータのON/OFF、受信データを取得します。
        
        @param id PMXサーボモータのID番号
        @param receiveMode PMXで設定した応答モード
        @param controlMode モータに設定する制御モード(ControlModeを参照)

        @retval [status, torqueSwitch, rxDataList]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval torqueSwitch FreeやTorqueOnの値(TorqueSwitchTypeを参照)
        @retval rxDataList MotorReadで読み取ったデータ[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧](合計8項目)
        """

        return self.MotorWRITE(id, PmxLib.TorqueSwitchType.Free, controlMode=controlMode, receiveMode=receiveMode)



    def setMotorBrake(self, id: int, receiveMode: int, controlMode=0x00):
        """
        指定した ID のモータをブレーキにし、ステータス、モータのON/OFF、受信データを取得します。
        
        @param id PMXサーボモータのID番号
        @param receiveMode PMXで設定した応答モード
        @param controlMode モータに設定する制御モード(ControlModeを参照)

        @retval [status, torqueSwitch, rxDataList]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval torqueSwitch FreeやTorqueOnの値(TorqueSwitchTypeを参照) 
        @retval rxDataList MotorReadで読み取ったデータ[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧](合計8項目)
        """
        

        return self.MotorWRITE(id, PmxLib.TorqueSwitchType.Brake, controlMode=controlMode, receiveMode=receiveMode)



    def setMotorHold(self, id: int, receiveMode: int, controlMode=0x00):
        """
        指定した ID のモータをホールドにし、ステータス、モータのON/OFF、受信データを取得します。
        
        @param id PMXサーボモータのID番号
        @param receiveMode PMXで設定した応答モード
        @param controlMode モータに設定する制御モード(ControlModeを参照)

        @retval [status, torqueSwitch, rxDataList]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval torqueSwitch FreeやTorqueOnの値(TorqueSwitchTypeを参照)
        @retval rxDataList MotorReadで読み取ったデータ[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧](合計8項目)
        """
        


        return self.MotorWRITE(id, PmxLib.TorqueSwitchType.Hold, controlMode=controlMode, receiveMode=receiveMode)



    def setPosition(self, id: int, receiveMode: int, position: int, speed=None, current=None, torque=None, time=None):
        """
        指定した ID のモータの位置、速度、電流、トルク、および時間パラメーターを設定し、データをモーター コントローラーに送信します。
        
        @param id PMXサーボモータのID番号
        @param receiveMode 応答モード。返ってきたデータを変換するときに使用します。
        @param position 位置パラメータは、モーターの望ましい位置を表す整数です。
        @param speed 速度制御を含む場合、速度値を引数として渡す
        @param current 電流制御を含む場合、電流値を引数として渡す
        @param torque トルク制御を含む場合、トルク値を引数として渡す
        @param time 時間制御を含む場合、補間時間を引数として渡す
        
        @retval [status, torqueSwitch, rxDataList]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval torqueSwitch FreeやTorqueOnの値(TorqueSwitchTypeを参照)
        @retval rxDataList MotorReadで読み取ったデータ[位置,速度,電流,トルク,PWM,モータ温度,CPU温度,電圧](合計8項目)
        
        @note MotorWRITE関数の返値を直接返します。

        @warning controlModeやreceiveModeは事前に設定してください。
        @warning controlModeで設定した目標値のみを引数に渡してください。
        @warning 複数組み合わせの場合で１つだけ変えるときでも全ての目標値を送ってください。
        
        @attention rxDataListでデータがない場合はリストの中身の各データがNoneになります
        
        """
        
        sendDataList = [position]        
        
        cMode = PmxLib.ControlMode.Position
        
        if speed is not None:
            sendDataList.append(speed)
            cMode += PmxLib.ControlMode.Current
        if current is not None:
            sendDataList.append(current)
            cMode += PmxLib.ControlMode.Current
        if torque is not None:
            sendDataList.append(torque)
            cMode += PmxLib.ControlMode.Torque
        if time is not None:
            sendDataList.append(time)
            cMode += PmxLib.ControlMode.Time
        
        return self.MotorWRITE(id, PmxLib.TorqueSwitchType.Control, controlMode=cMode, receiveMode=receiveMode, sendDataList=sendDataList)

    
    #################################################################################################################    
    
    #################################################################################################################
    
    # SystemWriteを利用したデータのget関数一覧
    
    #################################################################################################################


    def getSerialNumber(self, id: int):
        """
        指定したIDのサーボからシリアル番号を取得します。
        
        @param id PMXサーボモータのID番号

        @retval [status, serialNumData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval serialNumData シリアル番号。通信が失敗した場合はNoneが返ります。
        """

        status, sysReadDatas = self.SystemREAD(id)

        if status == PmxLib.ComError.OK:
            return status, sysReadDatas[0]
        else:
            return status, None

    def getModelNum(self, id: int):
        """
        指定したIDのサーボからモデル番号を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, modelNumData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval modelNumData モデル番号。通信が失敗した場合はNoneが返ります。
        """

        status, sysReadDatas = self.SystemREAD(id)

        if status == PmxLib.ComError.OK:
            return status, sysReadDatas[1]
            
        else:
            return status, None

    def getVersion(self, id: int):
        """
        指定したIDからファームウェアのバージョン番号を取得します。

        @param id PMXサーボモータのID番号
        
        @retval [status, varData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval varData バージョン番号のタプル[MAJOR,MINOR,PATCH,BUILD]。通信が失敗した場合はNoneが返ります
        
        """

        status, sysReadDatas = self.SystemREAD(id)

        if status == PmxLib.ComError.OK:
            return status, sysReadDatas[2]
            
        else:
            return status, None

    def getResponseTime(self, id: int):
        """
        システム読み取り操作のステータスと応答時間を取得します。
        
        @param id PMXサーボモータのID番号
        
        @retval [status, respData]
        @retval status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        @retval respData 応答時間[us]。通信が失敗した場合はNoneが返ります。
        
        """

        status, sysReadDatas = self.SystemREAD(id)

        if status == PmxLib.ComError.OK:
            return status, sysReadDatas[3]
            
        else:
            return status, None

    
    #################################################################################################################    
    
    #################################################################################################################
    # SystemWriteを利用したデータのset関数一覧
    #################################################################################################################
    
    def setId(self, id: int, newId: int, serialNum=None):
        """
        指定したIDを新しい ID に設定します。
        
        @param id PMXサーボモータの現在のID番号
        @param newId PMXサーボモータのWrite新しいID番号
        @param serialNum シリアル番号 わからない場合は SystemWRITEで取得するので省略可
        
        @return status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        
        @attention シリアル番号がNoneの場合はSystemWRITE関数内で読み込みます。
        """

        
        return self.SystemWRITE(id, serialNum=serialNum, newId=newId)



    def setBaudrate(self, id: int, newBaudrate: int, serialNum=None):
        """
        指定したIDの通信速度を設定します
        
        @param id PMXサーボモータのID番号
        @param newBaudrate 通信速度
        @param serialNum シリアル番号 わからない場合は SystemWRITEで取得するので省略可
        
        @return status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
                
        @attention シリアル番号がNoneの場合はSystemWRITE関数内で読み込みます。
        @attention 通信速度は関数内でPMXの通信速度値に変換するので通信速度は直接渡してください。
    
    
        """

        baudrateVal = 0
        
        if newBaudrate == 57600:
            baudrateVal = PmxLib.EditBaudrate._57600
        elif newBaudrate == 115200:
            baudrateVal = PmxLib.EditBaudrate._115200
        elif newBaudrate == 625000:
            baudrateVal = PmxLib.EditBaudrate._625000
        elif newBaudrate == 1000000:
            baudrateVal = PmxLib.EditBaudrate._1000000
        elif newBaudrate == 1250000:
            baudrateVal = PmxLib.EditBaudrate._1250000
        elif newBaudrate == 1500000:
            baudrateVal = PmxLib.EditBaudrate._1500000
        elif newBaudrate == 2000000:
            baudrateVal = PmxLib.EditBaudrate._2000000
        elif newBaudrate == 3000000:
            baudrateVal = PmxLib.EditBaudrate._3000000
        else:  # 上記以外は設定できないのでエラーを返す
            return PmxLib.ComError.FormatError
        
        return self.SystemWRITE(id, serialNum=serialNum, newBaudrateVal=baudrateVal)



    def setParity(self, id: int, newParity: int, serialNum=None):
        """
        指定したIDのパリティを設定します
        
        @param id PMXサーボモータのID番号
        @param newParity 通信で使用するパリティの設定(EditParityを参照)   
        @param serialNum シリアル番号 わからない場合は SystemWRITEで取得するので省略可
        @return status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        
        @attention シリアル番号がNoneの場合はSystemWRITE関数内で読み込みます。
        
        """

        return self.SystemWRITE(id, serialNum=serialNum, newParityVal=newParity)



    def setResponseTime(self, id: int, newResponseTime: int, serialNum=None):
        """
        指定したIDの応答時間を設定します
        
        @details  サーボ自身がデータを受信してから送信するまでのタイミング(ResponseTime)を変更します
        
        @param id PMXサーボモータのID番号
        @param newResponseTime 応答時間[us]
        @param serialNum シリアル番号 わからない場合は SystemWRITEで取得するので省略可
        
        @return status (Uint16) 通信の状態とステータス(送信結果(PmxStatusErrorList参照) + PMXのstatus)
        
        @attention シリアル番号がNoneの場合はSystemWRITE関数内で読み込みます。
        
        """

        return self.SystemWRITE(id, serialNum=serialNum, newResponseTime=newResponseTime)



#################################################################################################################


#################################################################################################################
#
# PMXのCRCの計算クラス
#
#################################################################################################################



class CRC16Class:
    """
    @brief PMXで使用するCRCの演算
    @details
     * 生成多項式：x^16+x^12+x^5+1 (CRC-16-CCITT)
     * ビットシフト方向：左送り
     * 初期値：0x0000
     * 出力：非反転 (出力XOR：0x0000)
    """

    ## PMXで使用するCRCのテーブル
    __crc_table = [
       0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
       0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
       0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
       0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
       0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
       0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
       0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
       0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
       0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
       0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
       0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
       0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
       0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
       0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
       0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
       0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
       0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
       0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
       0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
       0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
       0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
       0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
       0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
       0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
       0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
       0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
       0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
       0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
       0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
       0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
       0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
       0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0
    ]

    @classmethod
    def getCrc16(cls, buff):
        """
        PMXのコマンドの送信及び受信の CRC16 チェックサムを計算します。
        
        @param buff 送信および受信データの配列
        @return 計算された CRC16 値。
        
        @warning CRC16はbuff内に入れないでください。
        
        """
        
        crc = 0
        temp = 0
        for count in range(len(buff)):
            temp = ((buff[count] ^ (crc >> 8)) & 0xff) & 0xffff
            crc = (cls.__crc_table[temp] ^ (crc << 8)) & 0xffff
        return (crc ^ 0x0000) & 0xffff

    @classmethod
    def checkCrc16(cls, buff):
        """
        受信されたデータのCRCをチェックします
        
        @param buff 受信したデータの配列
        
        @return CRC16 値と一致する場合、True を返します。それ以外の場合は、False を返します。
        
        @warning 受信データにはCRCを含み受信データ全てを渡してください。
        
        """
        
        bufLength = len(buff)
        rcrc = int.from_bytes(buff[bufLength - 2: bufLength], byteorder='little')
        crc2 = cls.getCrc16(buff[0:bufLength - 2])
        if rcrc == crc2:
            return True
        else:
            return False
