#coding: UTF-8

import sys
import serial

from PmxLib import PmxLib
#import PmxLib


class PmxPySerialClass(PmxLib):
    """
    @brief PMXで使用する関数や定義をひとまとまりにしたもので、PySerialを使用して動作します。
    @details    ・PMXの固有値の定義
                ・PMXのコマンドの生成、チェックおよび送受信(送信部分は外部依存)
                ・PMXで必要だと思われる機能の関数化
                ・通信部分はPySerialを使用しています
                をしています
    @attention PySerialは別途インポートしてください。

    """
    
    
    _isSynchronize = False    
    _com = 0
    
    _logOutput = False

    """_summary_
    """
    def __init__(self):
        self._isSynchronize = False 

    @property
    def logOutput(self):
        return self._logOutput 

    @logOutput.setter
    def logOutput(self, value):
        """
        通信のログを16進数で出力する設定をする
        
        @param value True:ログ出力,False:ログを出力しない
        """
        
        self._logOutput = value 


    def __synchronizeWrite(self, txBuf):
        
        if self._logOutput == True: 
            print("tx>>", list(map(hex, txBuf)))
        
        sendbuf = [256]  # 送信するためのバッファを用意しておく
        sendbuf.clear()        
        
        # データのコピーとbyteかどうかの判定
        for i in range(len(txBuf)):
            sendbuf.append(txBuf[i])
        
        # 送信をする
        self._com.flushInput()  # buff clr
        self._com.write(sendbuf)
        

    
    def synchronize(self, txBuf, rxLen):
        """
        PySerialを使用してtxBufのコマンドを送受信します
        
        @param txBuf 送信データ
        @param rxLen 受信データ数
        
        @return 受信データ
        """

        if self._isSynchronize == False:

            # 使用中でアクセスできないようにロックをかけておく
            self._isSynchronize = True
            
            # データを書く
            self.__synchronizeWrite(txBuf)

            # データを取得する
            rxbuf = self._com.read(rxLen)
            self._com.flushInput()  # buff clr

            # 使用後でロックを解除する
            self._isSynchronize = False
            
            if self._logOutput == True: 
                print("rx<<", list(map(hex, rxbuf)))
        
            return rxbuf
            
        else:
            rxbuf = []  # error
            return rxbuf
        
    
    def synchronizeVariableRead(self, txBuf):
        """
        PySerialを使用してtxBufのコマンドを送受信します
        返信データ数がわからない場合に使用します。
        
        @param txBuf 送信データ
        
        @return 受信データ
        
        @note データを読む際、6byte目にLengthがあるので、まずは6byte取得する
        @note Lengthを取得できたらそのデータをもとに全てのデータを取得する
        @note データが来ない場合もあるのでTimeoutでエラー処理も行う
        
        """
        
        if self._isSynchronize == False:
            
            # 使用中でアクセスできないようにロックをかけておく
            self._isSynchronize = True
            
            # データを書く
            self.__synchronizeWrite(txBuf)
            
            # 最小単位の6byteをよむ
            rxFarstBuf = self._com.read(6)
            
            # 最小単位を読んだけど来なかった
            if (len(rxFarstBuf) != 6):
                self._isSynchronize = False
                rxbuf = []  # error
                return rxbuf
            
            # 返ってきたデータの4byte目がサイズにあたる
            rxDataSize = rxFarstBuf[3]
            
            # 残りを読む
            rxNextBuf = self._com.read(rxDataSize - 6)
            
            self._com.flushInput()  # buff clr
            
            # 使用後でロックを解除する
            self._isSynchronize = False
            
            rxbuf = []
            
            rxbuf.extend(rxFarstBuf)
            rxbuf.extend(rxNextBuf)

            if self._logOutput == True: 
                print("rx<<", list(map(hex, rxbuf)))
            
            return rxbuf
        else:
            rxbuf = []  # error
            return rxbuf

    def synchronizeNoRead(self, txBuf):
        """
        ブロードキャストIDなどでデータの返事がないとわかっている時に送信をします。
        
        @param txBuf 送信データ
        
        @return True:送信成功、False:送信失敗
        """
        
        if self._isSynchronize == False:
            
            # 使用中でアクセスできないようにロックをかけておく
            self._isSynchronize = True

            # データを書く
            self.__synchronizeWrite(txBuf)

            # 使用後でロックを解除する
            self._isSynchronize = False
    
            return True
        
        else:
            
            return False

    def setSerialParameters(self, baudrate=None, parity=None, timeout=None):
        """
        PySerialで使用する通信速度、パリティ、タイムアウトなどのシリアル パラメータを設定します。
        
        @param baudrate 通信速度
        @param parity パリティ
        @param timeout タイムアウト(sec単位(float型))
        """

        if baudrate is not None:
            self._com.baudrate = baudrate
        
        if parity is None:
            pass
        elif parity == PmxLib.EditParity.ParityNone:
            self._com.parity = serial.PARITY_NONE
        elif parity == PmxLib.EditParity.Odd:
            self._com.parity = serial.PARITY_ODD
        elif parity == PmxLib.EditParity.Even:
            self._com.parity = serial.PARITY_EVEN
        else:
            raise NotImplementedError("Parity parameter is invalid. ")
        
        if timeout is not None:
            self._com.timeout = timeout
        
        

    # ////////////////////////////////////////////////////////////////////
    # //	Serialポートを設定をして開く
    # //
    
    def open(self, comName, baudrate=115200, parity=serial.PARITY_NONE, timeout=1.0):
        """
        @brief	SerialポートをPMXの設定に合わせて開く
        @param	comName	ポートの名前
        @param	baudrate	通信速度
        @param  parity  パリティ
        @param	timeout	受信タイムアウト
        @return True:Open成功　False:Open失敗
        @note	Serialポートを開く
                openできなかったらFalseを返す
        """
        
        print("TryOPEN")
        print(comName)

        if self._com == 0:

            try:
                self._com = serial.Serial(comName, baudrate, parity=parity, stopbits=1, timeout=timeout)
                return True
            
            except serial.SerialException:
                return False
            except:        
                return False
            
        else:
            return False

    # ////////////////////////////////////////////////////////////////////
    # //	Serialポートを閉じる
    # //

    def close(self):
        """
        開いている_comポートを閉じる
        
        @return 閉じるのが成功したかどうかの状態を返す
        """
        try:
            self._com.close()
            self._com = 0
            return True
        except:
            return False



# /////////////////////////////////////////////////////////////////////////////
# //	ここまで[EOF]
# /////////////////////////////////////////////////////////////////////////////
