#coding: UTF-8

#
#  @file MemREAD_Sample20231023.py
#  @brief MemREAD sample code 
#  @author Kondo Kagaku Co.,Ltd.
#  @date 2023/10/23
#

#   
#   MemREADSample_20231023.pyはPMXのMemREAD関数に関する関数を実行するサンプルコードです。
#   MemREADSample_20231023.py is a sample code that executes functions related to the MemREAD function of PMX.
#

import sys
import os
import time

# PmxLib import
# ライブラリのパスはフォルダの構成によって変わります
# 現在のスクリプトのディレクトリを取得
current_dir = os.path.dirname(__file__)
# 1つ上のフォルダを取得(sanpleのフォルダの同じディレクトリにPmxLibがあるため)
parent_dir = os.path.dirname(current_dir)
# Pmxのライブラリのパスを登録
pmxLibPath = parent_dir + '\PmxLib'
sys.path.append(pmxLibPath)

from PmxPySerialClass  import PmxPySerialClass  # pySerialでPMXを通信するクラスをつかるようにする
from PmxLib import PmxLib as PMX  # PMXの固有値が定義してある「PmxLib」をPMXという名前で使えるようにする


# クラスをインスタンス化
pmx = PmxPySerialClass()

# 通信の設定項目
# comName = "/dev/ttyUSB0"
comName = "COM5"
pmxBaudrate = 3000000
pmxTimeout = 1.0

ServoID = 0

"""Serial open"""
def serialPortOpen():
    ### COM open
    openFlag = pmx.open(comName, baudrate=pmxBaudrate, timeout=pmxTimeout)
    if openFlag == False:
        print("Serial not open. Exit program")
        exit()
    else:
        print("Serial Connect success")
        



""" Main処理 """
def main():

    # 通信ログを表示させる(表示：True,非表示：False)
    pmx.logOutput = False

    # COMを開く
    serialPortOpen()


    # MemRead関数を直接使ってステータス情報を呼び出す(返信データはbyte配列)
    # 300番地はサーボの状態を読み出すことができます。
    # 6bytes読み出すことで現在位置、現在速度、現在電流値を一気に読み出せます。ただし、動作していませんので現在速度は0、現在電流値は微量な数値になります。
    print('300番地(現在位置)から6byte読み出します(返り値はbytes型)')
    flag, dataArray = pmx.MemREAD(ServoID, PMX.RamAddrList.NowPosition, 6)  # id , Addr ,dataSize
    print(hex(flag), dataArray)
    
    # pmx.MemREAD()で読み出したデータはリトルエンディアンで分割されています。以下の関数で一つのデータに連結することができます。
    # 例として現在位置のみ連結します。
    if flag == 0:  # flagが0だった時正常にデータが返ってきている
        posArray = [dataArray[0], dataArray[1]]          # 現在位置のデータは300, 301番地のため[0], [1]が該当
        pos = int.from_bytes(posArray, byteorder="little", signed=True)
        print("現在位置", pos)


    # pmx.MemREAD()の場合は分割されたデータを受け取るため連結し直す必要がありますが、下記の関数は戻り値で連結したデータを受け取ることができます。
    # MemREADで様々な型でデータを取得することができます。詳しくは下記アドレスを参照してください
    # https://www.kondokagaku.jp/wp_mn/archives/1144#6

    # 各データの型は「メモリマップ一覧」を参照してください。
    # https://www.kondokagaku.jp/wp_mn/archives/1144#i-36

    # MemREAD関数を使ってを現在位置（単位[1/100度]）(2byte符号あり)を取得する
    print('2byte(signed)の位置データを取得します')
    flag, data = pmx.MemREADToInt16(ServoID, PMX.RamAddrList.NowPosition)  # id, Addr
    if flag == 0:  # flagが0だった時正常にデータが返ってきている
        print(hex(flag), data / 100, "[°]")
    else:
        print(hex(flag), "None", "[°]")
    
    # MemRead関数を使って電圧(単位[mV])(2byte符号なし)を取得する
    print('2byte(unsigned)の電圧データを取得します')
    flag, data = pmx.MemREADToUint16(ServoID, PMX.RamAddrList.InputVoltage)  # id, Addr
    
    if flag == 0:  # flagが0だった時正常にデータが返ってきている
        print(hex(flag), data / 1000, "[V]")
    else:
        print(hex(flag), "None", "[V]")


    # MemRead関数を使って位置制御のPゲインを取得する
    print('4byte(unsigned)の位置制御のPゲインを取得します')
    flag, data = pmx.MemREADToUint32(ServoID, PMX.RamAddrList.PositionKp)  # id, Addr
    print(hex(flag), data)


    # MemRead関数を使ってトルクスイッチの状態を取得する
    print('1byte(unsigned)のトルクスイッチの状態を取得します')
    # TorqueSwitch Data: 0x01=TorqueON, 0x02=Free, 0x04=Brake, 0x08=Hold
    flag, data = pmx.MemREADToByte(ServoID, PMX.RamAddrList.TorqueSwitch)  # id, Addr
    print(hex(flag), data)


    # ライブラリには専用の関数を用意している場合があります。
    # 下記はトルクスイッチの状態を読み出すための関数です。これにより、アドレスを指定する必要がなくなり、プログラムもすっきりします。

    # トルクスイッチの状態を取得する
    print('トルクスイッチの状態を読み出します')
    # TorqueSwitch Data: 0x01=TorqueON, 0x02=Free, 0x04=Brake, 0x08=Hold
    flag, data = pmx.getTorqueSwitch(ServoID)  # id
    print(hex(flag), data)

    # エラー情報を読み出し、エラーを解除します
    print('エラーステータス(400番地から6byte)からエラー情報を読み出します')
    print('エラー情報を読み出すと解除されます')
    status, fullStatus = pmx.getFullStatus(ServoID)  # fullStarus = [staData(400)、sysStaData(401)、motStaData(402)、ramStaData(404)]
    print(hex(status), fullStatus)


    pmx.close()  # COMを閉じます


""" main """
if __name__ == "__main__":
    main()
